# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

import logging
import random
from pathlib import Path

import pytest

EXAMPLE_DATASET_TYPE_DOCS = map(
    str,
    Path(__file__)
    .parent.parent.joinpath("docs", "config_samples", "dataset_types")
    .glob("**/*.yaml"),
)


# Documents that shouldn't be accepted as mapping docs.
INVALID_MAPPING_DOCS = map(str, Path(__file__).parent.parent.joinpath("docs").glob("*"))


def _dataset_type_count(index):
    with index._active_connection() as connection:
        return len(list(connection.get_all_products()))


def test_add_example_dataset_types(
    clirunner, index, default_metadata_type, eo3_product_paths, ext_eo3_mdt_path
) -> None:
    """
    Add example mapping docs, to ensure they're valid and up-to-date.

    We add them all to a single database to check for things like duplicate ids.
    """
    existing_mappings = _dataset_type_count(index)

    if index.supports_legacy:
        # Legacy EO test examples
        print(f"{existing_mappings} mappings")
        for mapping_path in EXAMPLE_DATASET_TYPE_DOCS:
            print(f"Adding mapping {mapping_path}")

            result = clirunner(["-v", "product", "add", mapping_path])
            assert result.exit_code == 0

            mappings_count = _dataset_type_count(index)
            assert mappings_count > existing_mappings, (
                "Mapping document was not added: " + str(mapping_path)
            )
            existing_mappings = mappings_count

        result = clirunner(["-v", "metadata", "show", "-f", "json", "eo"])
        assert result.exit_code == 0

    # EO3 test examples
    result = clirunner(["-v", "metadata", "add", ext_eo3_mdt_path])
    assert result.exit_code == 0

    for path in eo3_product_paths:
        result = clirunner(["-v", "product", "add", path])
        assert result.exit_code == 0

        mappings_count = _dataset_type_count(index)
        assert mappings_count > existing_mappings, (
            "Mapping document was not added: " + str(path)
        )
        existing_mappings = mappings_count

    result = clirunner(["-v", "metadata", "show", "-f", "json", "eo3"])
    assert result.exit_code == 0

    result = clirunner(["-v", "metadata", "list"])
    assert result.exit_code == 0

    result = clirunner(["-v", "metadata", "show"])
    assert result.exit_code == 0

    result = clirunner(["-v", "product", "list"])
    assert result.exit_code == 0

    expect_result = 0 if existing_mappings > 0 else 1
    result = clirunner(["-v", "product", "show"], expect_success=(expect_result == 0))
    assert result.exit_code == expect_result

    if existing_mappings > 1:
        result = clirunner(
            ["-v", "product", "show", "-f", "json"], expect_success=False
        )
        assert result.exit_code == 1

        result = clirunner(["-v", "product", "show", "-f", "json", "ga_ls8c_ard_3"])
        assert result.exit_code == 0

        result = clirunner(["-v", "product", "show", "-f", "yaml", "ga_ls8c_ard_3"])
        assert result.exit_code == 0


def test_error_returned_on_invalid(clirunner, index) -> None:
    assert _dataset_type_count(index) == 0

    for mapping_path in INVALID_MAPPING_DOCS:
        result = clirunner(
            ["product", "add", mapping_path],
            # TODO: Make this false when the cli is updated to print errors (rather than uncaught exceptions).
            catch_exceptions=True,
            expect_success=False,
        )
        assert result.exit_code != 0, "Success return code for invalid document."
        assert _dataset_type_count(index) == 0, "Invalid document was added to DB"


def test_config_check(clirunner, index, cfg_env) -> None:
    # This is not a very thorough check, we just check to see that
    # it prints something vaguely related and does not error-out.
    result = clirunner(["system", "check"])

    assert cfg_env["db_hostname"] in result.output
    assert cfg_env["db_username"] in result.output
    assert str(cfg_env["skip_broken_datasets"]) in result.output
    assert str(cfg_env["dc_load_limit"]) in result.output


def test_list_users_does_not_fail(clirunner, cfg_env, index) -> None:
    # We don't want to make assumptions about available users during test runs.
    # (They are host-global, not specific to the database)
    # So we're just checking that it doesn't fail (and the SQL etc is well formed)
    result = clirunner(["user", "list"])
    assert result.exit_code == 0


def test_db_init_noop(clirunner, cfg_env, ls8_eo3_product) -> None:
    # Run on an existing database.
    result = clirunner(["-v", "system", "init"])
    assert "Updated." in result.output
    # It should not rebuild indexes by default
    assert f"Dropping index: dix_{ls8_eo3_product.name}" not in result.output

    result = clirunner(["metadata", "list"])
    assert "eo3 " in result.output


@pytest.mark.parametrize("datacube_env_name", ("datacube",))
def test_db_init_rebuild(clirunner, cfg_env, ls5_telem_type) -> None:
    if cfg_env._name == "datacube":
        from datacube.drivers.postgres import _dynamic
        from datacube.drivers.postgres.sql import SCHEMA_NAME

        # Set field creation logging to debug since we assert on debug output.
        _dynamic._LOG.setLevel(logging.DEBUG)
    else:
        from datacube.drivers.postgis.sql import SCHEMA_NAME
    # Run on an existing database.
    result = clirunner(["-v", "-E", cfg_env._name, "system", "init", "--rebuild"])
    assert "Updated." in result.output
    # It should have recreated views and indexes.
    assert f"Dropping index: dix_{ls5_telem_type.name}" in result.output
    assert f"Creating index: dix_{ls5_telem_type.name}" in result.output
    assert (
        f"Dropping view: {SCHEMA_NAME}.dv_{ls5_telem_type.name}_dataset"
        in result.output
    )
    assert (
        f"Creating view: {SCHEMA_NAME}.dv_{ls5_telem_type.name}_dataset"
        in result.output
    )


def test_db_init(clirunner, index) -> None:
    if index._db.driver_name == "postgis":
        from datacube.drivers.postgis._core import drop_db, has_schema
    else:
        from datacube.drivers.postgres._core import drop_db, has_schema

    with index._db._connect() as connection:
        drop_db(connection._connection)

        assert not has_schema(index._db._engine)

    # Run on an empty database.
    if index._db.driver_name == "postgis":
        result = clirunner(["-E", "postgis", "system", "init"])
    else:
        result = clirunner(["system", "init"])

    assert "Created." in result.output

    with index._db._connect() as connection:
        assert has_schema(index._db._engine)


def test_add_no_such_product(clirunner, index) -> None:
    result = clirunner(
        ["dataset", "add", "--dtype", "no_such_product", "/tmp"], expect_success=False
    )
    assert result.exit_code != 0
    assert "DEPRECATED option detected" in result.output
    assert "ERROR Supplied product name" in result.output


@pytest.fixture(
    params=[
        ('test_"user"_{n}', None),
        ('test_"user"_{n}', "Test user description"),
        # Test that names are escaped
        ('test_user_"invalid+_chars_{n}', None),
        ("test_user_invalid_desc_{n}", "Invalid \"' chars in description"),
    ]
)
def example_user(clirunner, index, request):
    username, description = request.param

    username = username.format(n=random.randint(111111, 999999))

    # test_roles = (user_name for role_name, user_name, desc in roles if user_name.startswith('test_'))
    with index._db._connect() as connection:
        users = (user_name for role_name, user_name, desc in connection.list_users())
        if username in users:
            connection.drop_users([username])

    # No user exists.
    assert_no_user(clirunner, username)

    yield username, description

    with index._db._connect() as connection:
        users = (user_name for role_name, user_name, desc in connection.list_users())
        if username in users:
            connection.drop_users([username])


def test_user_creation(clirunner, example_user) -> None:
    """
    Add a user, grant them, delete them.

    This test requires role creation privileges on the PostgreSQL instance used for testing...
    """
    username, user_description = example_user

    # Create them
    args = ["user", "create", "user", username]
    if user_description:
        args.extend(["--description", user_description])
    clirunner(args)
    assert_user_with_role(clirunner, "ingest", username)

    # Grant them 'manage' permission
    clirunner(["user", "grant", "manage", username])
    assert_user_with_role(clirunner, "manage", username)

    # Delete them
    clirunner(["user", "delete", username])
    assert_no_user(clirunner, username)


def assert_user_with_role(clirunner, role, user_name) -> None:
    result = clirunner(["user", "list"])
    assert "{}{}".format("user: ", user_name) in result.output


def assert_no_user(clirunner, username) -> None:
    result = clirunner(["user", "list"])
    assert username not in result.output



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""Non-db specific implementation of metadata search fields.

This allows extraction of fields of interest from dataset metadata document.
"""

import decimal
from collections.abc import Callable, Mapping
from typing import Any

import toolz
from typing_extensions import override

from datacube.utils import parse_time

from ._base import Range

# Allowed values for field 'type' (specified in a metadata type document)
_AVAILABLE_TYPE_NAMES = (
    # Unrestricted type - handy for dynamically creating fields from offsets, e.g. for search_returning()
    "any",
    "numeric-range",
    "double-range",
    "integer-range",
    "datetime-range",
    "string",
    "numeric",
    "double",
    "integer",
    "boolean",
    "datetime",
    # For backwards compatibility (alias for numeric-range)
    "float-range",
)


class Expression:
    # No properties at the moment. These are built and returned by the
    # DB driver (from Field methods), so they're mostly an opaque token.

    # A simple equals implementation for comparison in test code.
    @override
    def __eq__(self, other) -> bool:
        if self.__class__ != other.__class__:
            return False
        return self.__dict__ == other.__dict__

    def evaluate(self, ctx):
        raise NotImplementedError()


class SimpleEqualsExpression(Expression):
    def __init__(self, field, value) -> None:
        self.field = field
        self.value = value

    @override
    def evaluate(self, ctx):
        return self.field.extract(ctx) == self.value


class Field:
    """
    A searchable field within a dataset/storage metadata document.
    """

    # type of field.
    # If type is not specified, the field is a string
    # This should always be one of _AVAILABLE_TYPE_NAMES
    type_name = "string"

    def __init__(self, name: str, description: str) -> None:
        self.name = name

        self.description = description

        # Does selecting this affect the output rows?
        # (eg. Does this join other tables that aren't 1:1 with datasets.)
        self.affects_row_selection = False

        assert self.type_name in _AVAILABLE_TYPE_NAMES, (
            f"Invalid type name {self.type_name!r}"
        )

    @override
    def __eq__(self, value) -> Expression:  # type: ignore[override]
        """
        Is this field equal to a value?

        this returns an Expression object (hence type ignore above)
        """
        raise NotImplementedError("equals expression")

    def between(self, low, high) -> Expression:
        """
        Is this field in a range?
        """
        raise NotImplementedError("between expression")

    # Should be True if value can be extracted from a dataset metadata document with the extract method
    can_extract: bool = False

    def extract(self, doc):
        raise NotImplementedError(f"extract for {self.name}")


class SimpleField(Field):
    def __init__(
        self,
        offset: list[str | int],
        converter: type | Callable[[Any], Any],
        type_name: str,
        name: str = "",
        description: str = "",
    ) -> None:
        self._offset = offset
        self._converter = converter
        self.type_name = type_name
        super().__init__(name, description)

    @override
    def __eq__(self, value) -> Expression:  # type: ignore[override]
        return SimpleEqualsExpression(self, value)

    can_extract = True

    @override
    def extract(self, doc):
        v = toolz.get_in(self._offset, doc, default=None)
        if v is None:
            return None
        return self._converter(v)


class RangeField(Field):
    def __init__(
        self,
        min_offset,
        max_offset,
        base_converter,
        type_name: str,
        name: str = "",
        description: str = "",
    ) -> None:
        self.type_name = type_name
        self._converter = base_converter
        self._min_offset = min_offset
        self._max_offset = max_offset
        super().__init__(name, description)

    can_extract = True

    @override
    def extract(self, doc) -> Range | None:
        def extract_raw(paths):
            vv = [toolz.get_in(p, doc, default=None) for p in paths]
            return [self._converter(v) for v in vv if v is not None]

        v_min = extract_raw(self._min_offset)
        v_max = extract_raw(self._max_offset)

        v_min = None if len(v_min) == 0 else min(v_min)
        v_max = None if len(v_max) == 0 else max(v_max)

        if v_min is None and v_max is None:
            return None

        return Range(v_min, v_max)


def parse_search_field(doc, name: str = "") -> RangeField | SimpleField:
    parsers: dict[str, type | Callable[[Any], Any]] = {
        "string": str,
        "double": float,
        "integer": int,
        "boolean": bool,
        "numeric": decimal.Decimal,
        "datetime": parse_time,
        "object": lambda x: x,
    }
    _type = doc.get("type", "string")

    if _type in parsers:
        offset = doc.get("offset", None)
        if offset is None:
            raise ValueError("Missing offset")

        return SimpleField(
            offset,
            parsers[_type],
            _type,
            name=name,
            description=doc.get("description", ""),
        )

    if not _type.endswith("-range"):
        raise ValueError("Unsupported search field type: " + str(_type))

    raw_type = _type.split("-")[0]

    if (
        raw_type == "float"
    ):  # float-range is supposed to be supported, but not just float?
        raw_type = "numeric"
        _type = "numeric-range"

    if raw_type not in parsers:
        raise ValueError("Unsupported search field type: " + str(_type))

    min_offset = doc.get("min_offset", None)
    max_offset = doc.get("max_offset", None)

    if min_offset is None or max_offset is None:
        raise ValueError("Need to specify both min_offset and max_offset")

    return RangeField(
        min_offset,
        max_offset,
        parsers[raw_type],
        _type,
        name=name,
        description=doc.get("description", ""),
    )


def get_dataset_fields(metadata_definition: Mapping[str, Any]) -> dict[str, Field]:
    """Construct search fields dictionary not tied to any specific db
    implementation.
    """
    fields = toolz.get_in(["dataset", "search_fields"], metadata_definition, {})
    return {n: parse_search_field(doc, name=n) for n, doc in fields.items()}



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import csv
import io
from collections.abc import Iterable

import datacube.scripts.search_tool
from datacube.model import Dataset, Product


def _load_product_query(
    lazy_results: Iterable[tuple[Product, Iterable[Dataset]]],
) -> dict[str, list[Dataset]]:
    """
    search_by_product() returns two levels of laziness. load them all into memory
    for easy comparison/counts
    """
    products: dict[str, list[Dataset]] = {}
    for product, datasets in lazy_results:
        assert product.name not in products, (
            "search_by_product() returned a product twice"
        )
        products[product.name] = list(datasets)
    return products


def _csv_search_raw(args, clirunner) -> str:
    # Do a CSV search from the cli, returning output as a string
    result = clirunner(
        ["-f", "csv"] + list(args),
        cli_method=datacube.scripts.search_tool.cli,
        verbose_flag=False,
    )
    output = result.output
    output_lines = output.split("\n")
    return "\n".join(line for line in output_lines if "WARNING" not in line)


def _cli_csv_search(args, clirunner) -> list:
    # Do a CSV search from the cli, returning results as a list of dictionaries
    output = _csv_search_raw(args, clirunner)
    return list(csv.DictReader(io.StringIO(output)))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from collections.abc import Callable
from typing import Any
from urllib.parse import urlparse

from deprecat import deprecat

from datacube.model import Dataset
from datacube.utils.uris import pick_uri, uri_resolve


def _get_band_and_layer(b: dict[str, Any]) -> tuple[int | None, str | None]:
    """Encode legacy logic for extracting band/layer:

        on input:
        band -- Int | Nothing
        layer -- Str | Int | Nothing

    Valid combinations are:
        band  layer  Output
    ---------------------------
          -     -    ( - ,  - )
          -    int   (int,  - )
         int    -    (int,  - )
         int   str   (int, str)
          -    str   ( - , str)
    """
    band = b.get("band")
    layer = b.get("layer")

    if band is None:
        if isinstance(layer, int):
            return layer, None
        if layer is None or isinstance(layer, str):
            return None, layer

        raise ValueError(
            f"Expect `layer` to be one of None,int,str but it is {type(layer)}"
        )
    else:
        if not isinstance(band, int):
            raise ValueError(f"Expect `band` to be an integer (it is {type(band)})")
        if layer is not None and not isinstance(layer, str):
            raise ValueError(
                f"Expect `layer` to be one of None,str but it is {type(layer)}"
            )

        return band, layer


def _extract_driver_data(ds: Dataset, mm: dict[str, Any]) -> Any | None:
    ds_data = ds.metadata_doc.get("driver_data", None)
    mm_data = mm.get("driver_data")
    if isinstance(ds_data, dict) and isinstance(mm_data, str):
        return ds_data.get(mm_data, {})
    if mm_data is not None:
        return mm_data
    return ds_data


def measurement_paths(ds: Dataset) -> dict[str, str]:
    """
    Returns a dictionary mapping from band name to url pointing to band storage
    resource.

    :return: Band Name => URL
    """
    if not ds.uri:
        raise ValueError("No locations on this dataset")
    base = pick_uri(ds.uris) if ds.has_multiple_uris() else ds.uri

    return {k: uri_resolve(base, m.get("path")) for k, m in ds.measurements.items()}


class BandInfo:
    __slots__ = (
        "band",
        "crs",
        "dims",
        "driver_data",
        "dtype",
        "format",
        "layer",
        "name",
        "nodata",
        "transform",
        "units",
        "uri",
    )

    @deprecat(
        deprecated_args={
            "extra_dim_index": {
                "version": "1.9",
                "reason": "3D code is deprecated",
            }
        }
    )
    def __init__(
        self,
        ds: Dataset,
        band: str,
        uri_scheme: str | None = None,
        extra_dim_index: int | None = None,
        patch_url: Callable[[str], str] | None = None,
    ) -> None:
        try:
            (mp,) = ds.product.lookup_measurements([band]).values()
        except KeyError:
            raise ValueError(f"No such band: {band}") from None

        mm = ds.measurements.get(mp.canonical_name)

        if mm is None:
            raise ValueError(f"No such band: {band}")

        if not ds.uri:
            raise ValueError("No uris defined on a dataset")
        base_uri = ds.legacy_uri(uri_scheme) if ds.has_multiple_uris() else ds.uri
        uri = uri_resolve(base_uri, mm.get("path"))
        if patch_url is not None:
            uri = patch_url(uri)

        bint, layer = _get_band_and_layer(mm)
        self.uri = uri
        self.name = band
        self.band = bint
        self.layer = layer
        self.dtype = mp.dtype
        self.nodata = mp.nodata
        self.units = mp.units
        self.crs = ds.crs
        self.transform = ds.transform
        self.format = ds.format or ""
        self.dims = mp.get("dims", None)
        self.driver_data = _extract_driver_data(ds, mm)

    @property
    def uri_scheme(self) -> str:
        return urlparse(self.uri).scheme or ""



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import os
from unittest import mock

import moto
import pytest

from datacube.testutils import write_files
from datacube.utils.aws import configure_s3_access
from datacube.utils.rio import (
    activate_from_config,
    activate_rio_env,
    deactivate_rio_env,
    get_rio_env,
    set_default_rio_config,
)


def test_rio_env_no_aws() -> None:
    deactivate_rio_env()

    # make sure we start without env configured
    assert get_rio_env() == {}

    ee = activate_rio_env(FAKE_OPTION=1)
    assert ee == get_rio_env()
    assert "GDAL_DISABLE_READDIR_ON_OPEN" not in ee
    if os.getenv("GDAL_DATA", None):
        assert "GDAL_DATA" in ee
    assert ee.get("FAKE_OPTION") == 1
    assert "AWS_ACCESS_KEY_ID" not in ee

    ee = activate_rio_env(cloud_defaults=True)
    assert "GDAL_DISABLE_READDIR_ON_OPEN" in ee
    assert ee == get_rio_env()

    deactivate_rio_env()
    assert get_rio_env() == {}


def test_rio_env_aws() -> None:
    deactivate_rio_env()

    # make sure we start without env configured
    assert get_rio_env() == {}

    with pytest.raises(ValueError):
        activate_rio_env(aws="something")  # type: ignore[arg-type]

    # note: setting region_name to avoid auto-lookup
    ee = activate_rio_env(aws={"aws_unsigned": True, "region_name": "us-west-1"})

    assert ee.get("AWS_NO_SIGN_REQUEST") == "YES"

    ee = activate_rio_env(
        cloud_defaults=True,
        aws={
            "aws_secret_access_key": "blabla",
            "aws_access_key_id": "not a real one",
            "aws_session_token": "faketoo",
            "region_name": "us-west-1",
        },
    )

    assert "AWS_NO_SIGN_REQUEST" not in ee
    # check secrets are sanitized
    assert ee.get("AWS_ACCESS_KEY_ID") == "xx..xx"
    assert ee.get("AWS_SECRET_ACCESS_KEY") == "xx..xx"
    assert ee.get("AWS_SESSION_TOKEN") == "xx..xx"

    assert ee.get("AWS_REGION") == "us-west-1"

    # check sanitize can be turned off
    ee = get_rio_env(sanitize=False)
    assert ee.get("AWS_SECRET_ACCESS_KEY") == "blabla"
    assert ee.get("AWS_ACCESS_KEY_ID") == "not a real one"
    assert ee.get("AWS_SESSION_TOKEN") == "faketoo"

    deactivate_rio_env()
    assert get_rio_env() == {}


def test_rio_env_aws_auto_region(monkeypatch, without_aws_env) -> None:
    import datacube.utils.aws

    pp = write_files(
        {
            "config": """[default]
"""
        }
    )

    assert (pp / "config").exists()
    monkeypatch.setenv("AWS_CONFIG_FILE", str(pp / "config"))

    assert datacube.utils.aws.botocore_default_region() is None

    aws = {
        "aws_secret_access_key": "blabla",
        "aws_access_key_id": "not a real one",
        "aws_session_token": "faketoo",
    }

    with mock.patch("datacube.utils.aws.ec2_current_region", return_value="TT"):
        ee = activate_rio_env(aws=aws)
        assert ee.get("AWS_REGION") == "TT"

    with mock.patch("datacube.utils.aws.ec2_current_region", return_value=None):
        ee = activate_rio_env(aws=aws)
        assert "AWS_REGION" not in ee

        with pytest.raises(ValueError):
            activate_rio_env(aws={"region_name": "auto"})

    deactivate_rio_env()
    assert get_rio_env() == {}


def test_rio_env_aws_auto_region_dummy() -> None:
    """Just call it we don't know if it will succeed"""

    with moto.mock_aws():
        # at least it should not raise error since we haven't asked for region_name='auto'
        _ = activate_rio_env(aws={})

        deactivate_rio_env()
        assert get_rio_env() == {}


def test_rio_env_via_config() -> None:
    ee = activate_from_config()
    assert ee is not None

    # Second call should not change anything
    assert activate_from_config() is None

    set_default_rio_config(aws=None, cloud_defaults=True)

    # config change should activate new env
    ee = activate_from_config()
    assert ee is not None
    assert "GDAL_DISABLE_READDIR_ON_OPEN" in ee

    deactivate_rio_env()
    assert get_rio_env() == {}


def test_rio_configure_aws_access(monkeypatch, without_aws_env, dask_client) -> None:
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "fake-key-id")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "fake-secret")
    monkeypatch.setenv("AWS_DEFAULT_REGION", "fake-region")

    creds = configure_s3_access()
    assert creds is not None
    cc = creds.get_frozen_credentials()
    assert cc.access_key == "fake-key-id"
    assert cc.secret_key == "fake-secret"
    assert cc.token is None

    ee = activate_from_config()
    assert ee is not None
    assert "AWS_ACCESS_KEY_ID" in ee
    assert "AWS_SECRET_ACCESS_KEY" in ee
    assert "AWS_REGION" in ee
    assert "AWS_SESSION_TOKEN" not in ee

    ee = get_rio_env(sanitize=False)
    assert ee is not None
    assert ee["AWS_ACCESS_KEY_ID"] == "fake-key-id"
    assert ee["AWS_SECRET_ACCESS_KEY"] == "fake-secret"
    assert ee["AWS_REGION"] == "fake-region"
    assert ee["GDAL_DISABLE_READDIR_ON_OPEN"] == "EMPTY_DIR"

    ee_local = ee
    client = dask_client

    creds = configure_s3_access(client=client)
    assert creds is not None
    cc = creds.get_frozen_credentials()
    assert cc.access_key == "fake-key-id"
    assert cc.secret_key == "fake-secret"
    assert cc.token is None

    ee = client.submit(activate_from_config).result()
    assert ee is not None
    assert "AWS_ACCESS_KEY_ID" in ee
    assert "AWS_SECRET_ACCESS_KEY" in ee
    assert "AWS_REGION" in ee
    assert "AWS_SESSION_TOKEN" not in ee

    def _activate_and_get(sanitize=True):
        activate_from_config()
        return get_rio_env(sanitize=sanitize)

    ee = client.submit(_activate_and_get, sanitize=False).result()
    assert ee == ee_local



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
import sys
from collections.abc import Sequence

import click
from click import confirm, echo
from odc.geo import CRS
from pyproj.exceptions import CRSError

from datacube.index import Index
from datacube.ui import click as ui
from datacube.ui.click import cli, print_help_msg

_LOG: logging.Logger = logging.getLogger("datacube-system")


@cli.group(name="spindex", help="Spatial Indexes")
def system() -> None:
    pass


@system.command("create", help="Create unpopulated spatial index for particular CRSes")
@click.option(
    "--update/--no-update",
    "-u",
    is_flag=True,
    default=False,
    help="Populate the spatial index after creation (slow). "
    "For finer grained updating, use the 'spindex update' command",
)
@click.argument("srids", nargs=-1)
@ui.pass_index()
def create(index: Index, update: bool, srids: Sequence[str]) -> None:
    if not index.supports_spatial_indexes:
        echo(
            "The active index driver does not support spatial indexes. "
            "You may want to specify a different environment or config (see --help for options)",
            err=True,
        )
        exit(1)
    if not srids:
        echo("Must supply at least one CRS to create/update", err=True)
        print_help_msg(create)
        exit(1)

    confirmed = []
    failed = []
    for srid in srids:
        try:
            crs = CRS(srid)
        except CRSError:
            failed.append(srid)
            echo(f"{srid} cannot be converted to a valid CRS")
            continue
        if not crs.epsg:
            failed.append(srid)
            echo(f"{srid} cannot be converted to a CRS with an EPSG number")
            continue
        if crs in index.spatial_indexes():
            # A spatial index for crs already exists: skip silently
            confirmed.append(crs)
        elif index.create_spatial_index(crs):
            # Creation succeeded
            confirmed.append(crs)
        else:
            # Creation attempted but failed
            failed.append(srid)
    if failed:
        str_failed = ",".join(f"{srid}" for srid in failed)
        echo(f"Could not create spatial indexes for: {str_failed}")
    if confirmed:
        echo(
            f"Spatial indexes created for: {','.join(str(crs.epsg) for crs in confirmed)}"
        )
    if update and failed:
        echo("Skipping update")
    elif update:
        result = index.update_spatial_index(confirmed)
        echo(f"{result} extents checked and updated in spatial indexes")
    else:
        echo(
            "Newly created spatial indexes are unpopulated - run the 'spindex update' command before use."
        )
    exit(len(failed))


@system.command(
    "list", help="List all CRSs for which spatial indexes exist in this index"
)
@ui.pass_index()
def list_spindex(index) -> None:
    if not index.supports_spatial_indexes:
        echo(
            "The active index driver does not support spatial indexes. "
            "You may want to specify a different environment or config (see --help for options)",
            err=True,
        )
        exit(1)
    for crs in index.spatial_indexes():
        echo(f"EPSG:{crs.epsg}")
    exit(0)


@system.command("update", help="Update a spatial index for particular CRSs.")
@click.option(
    "--product",
    "-p",
    multiple=True,
    help="The name of a product to update the spatial index for (can be used multiple times for multiple products)",
)
@click.option(
    "--dataset",
    "-d",
    multiple=True,
    help="The id of a dataset to update the spatial index for (can be used multiple times for multiple datasets)",
)
@click.argument("srids", nargs=-1)
@ui.pass_index()
def update(
    index: Index, product: Sequence[str], dataset: Sequence[str], srids: Sequence[str]
) -> None:
    if not index.supports_spatial_indexes:
        echo(
            "The active index driver does not support spatial indexes. "
            "You may want to specify a different environment or config (see --help for options)",
            err=True,
        )
        exit(1)
    if not srids:
        echo("Must supply at least one CRS to update", err=True)
        print_help_msg(update)
        exit(1)

    for_update = []
    cant_update = []
    for srid in srids:
        try:
            crs = CRS(srid)
        except CRSError:
            echo(f"{srid} is not a valid CRS: skipping")
            cant_update.append(srid)
            continue
        if not crs.epsg:
            echo(f"{srid} cannot be converted to a CRS with an EPSG number")
            cant_update.append(srid)
            continue

        if crs in index.spatial_indexes():
            for_update.append(crs)
        else:
            cant_update.append(srid)
            echo(f"No spatial index for crs {srid} exists: skipping")
    if not for_update:
        echo("Nothing to update!")
        exit(len(cant_update))
    result = index.update_spatial_index(
        for_update, product_names=product, dataset_ids=dataset
    )
    echo(f"{result} extents checked and updated in {len(for_update)} spatial indexes")
    exit(len(cant_update))


@system.command("drop", help="Drop existing spatial indexes for particular CRSs")
@click.option(
    "--force/--no-force",
    "-f",
    is_flag=True,
    default=False,
    help="If set, does not ask the user to confirm deletion",
)
@click.argument("srids", nargs=-1)
@ui.pass_index()
def drop(index: Index, force: bool, srids: Sequence[str]) -> None:
    if not index.supports_spatial_indexes:
        echo(
            "The active index driver does not support spatial indexes. "
            "You may want to specify a different environment or config (see --help for options)",
            err=True,
        )
        exit(1)
    if not srids:
        echo("Must supply at least one CRS to drop", err=True)
        print_help_msg(drop)
        exit(1)
    for_deletion = []
    errors = False
    for srid in srids:
        try:
            crs = CRS(srid)
        except CRSError:
            echo(f"{srid} is not a valid CRS: skipping")
            errors = True
            continue
        if not crs.epsg:
            echo(f"{srid} cannot be converted to a CRS with an EPSG number")
            errors = True
            continue
        if crs in index.spatial_indexes():
            for_deletion.append(crs)
        else:
            echo(f"No spatial index exists for CRS {srid} - skipping")
    if for_deletion and not force:
        echo(
            "WARNING: Recreating spatial indexes may be slow and expensive for large databases."
        )
        echo(
            "You have requested to delete spatial indexes for the following "
            f"CRSes: {','.join(str(crs.epsg) for crs in for_deletion)}"
        )
        if sys.stdin.isatty():
            confirmed = confirm(
                "Are you sure you want to delete these spatial indexes?", default=False
            )
            if not confirmed:
                echo("OK aborting", err=True)
                exit(1)
        else:
            echo("Use --force from non-interactive scripts. Aborting.")
            exit(1)
    for crs in for_deletion:
        click.echo(f"Deleting spatial index for CRS epsg:{crs.epsg}: ", nl=False)
        if index.drop_spatial_index(crs):
            click.echo("Done")
        else:
            click.echo("Failed")
            errors = True
    if errors:
        exit(1)
    else:
        exit(0)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from types import SimpleNamespace

import pytest
import yaml

from datacube.drivers import (
    index_driver_by_name,
    index_drivers,
    new_datasource,
    reader_drivers,
    writer_drivers,
)
from datacube.drivers.indexes import IndexDriverCache
from datacube.model import MetadataType
from datacube.storage import BandInfo
from datacube.storage._rio import RasterDatasetDataSource
from datacube.testutils import mk_sample_dataset, suppress_deprecations


def test_new_datasource_fallback() -> None:
    bands = [{"name": "green", "path": ""}]
    dataset = mk_sample_dataset(bands, "file:///foo", format="GeoTiff")

    assert dataset.uri_scheme == "file"

    rdr = new_datasource(BandInfo(dataset, "green"))
    assert rdr is not None
    assert isinstance(rdr, RasterDatasetDataSource)

    # check that None format works
    band = BandInfo(mk_sample_dataset(bands, "file:///file", format=None), "green")
    rdr = new_datasource(band)
    assert rdr is not None
    assert isinstance(rdr, RasterDatasetDataSource)


def test_reader_drivers() -> None:
    available_drivers = reader_drivers()
    assert isinstance(available_drivers, list)


def test_writer_drivers() -> None:
    available_drivers = writer_drivers()
    assert "netcdf" in available_drivers
    assert "NetCDF CF" in available_drivers


def test_index_drivers() -> None:
    available_drivers = index_drivers()
    assert "default" in available_drivers
    assert "null" in available_drivers
    assert "memory" in available_drivers


def test_default_injection() -> None:
    cache = IndexDriverCache("datacube.plugins.index-no-such-prefix")
    assert set(cache.drivers()) == {
        "default",
        "postgres",
        "legacy",
        "postgis",
        "memory",
    }


def test_netcdf_driver_import() -> None:
    try:
        import datacube.drivers.netcdf.driver
    except ImportError:
        raise Exception("Failed to load netcdf writer driver") from None

    assert datacube.drivers.netcdf.driver.reader_driver_init is not None


def test_writer_driver_mk_uri() -> None:
    from datacube.drivers.netcdf.driver import NetcdfWriterDriver

    writer_driver = NetcdfWriterDriver()

    assert writer_driver.uri_scheme == "file"

    file_path = "/path/to/my_file.nc"
    file_uri = writer_driver.mk_uri(file_path=file_path)
    assert file_uri == f"file://{file_path}"


def test_metadata_type_from_doc() -> None:
    metadata_doc = yaml.safe_load("""
name: minimal
description: minimal metadata definition
dataset:
    id: [id]
    sources: [lineage, source_datasets]
    label: [label]
    creation_dt: [creation_dt]
    search_fields:
        some_custom_field:
            description: some custom field
            offset: [a,b,c,custom]
    """)

    for name in index_drivers():
        driver = index_driver_by_name(name)
        assert driver is not None
        with suppress_deprecations():
            metadata = driver.metadata_type_from_doc(metadata_doc)  # deprecated method
            assert isinstance(metadata, MetadataType)
            assert metadata.id is None
            assert metadata.name == "minimal"
            assert "some_custom_field" in metadata.dataset_fields


def test_reader_cache_throws_on_missing_fallback() -> None:
    from datacube.drivers.readers import rdr_cache

    rdrs = rdr_cache()
    assert rdrs is not None

    with pytest.raises(KeyError):
        rdrs("file", "aint-such-format")


def test_driver_singleton() -> None:
    from unittest.mock import MagicMock

    from datacube.drivers._tools import singleton_setup

    result = object()
    factory = MagicMock(return_value=result)
    obj = SimpleNamespace()

    assert singleton_setup(obj, "xx", factory) is result
    assert singleton_setup(obj, "xx", factory) is result
    assert singleton_setup(obj, "xx", factory) is result
    assert obj.xx is result

    factory.assert_called_once_with()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from . import writer as netcdf_writer
from ._write import create_netcdf_storage_unit, write_dataset_to_netcdf
from .writer import Variable

__all__ = [
    "Variable",
    "create_netcdf_storage_unit",
    "netcdf_writer",
    "write_dataset_to_netcdf",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import json
from typing import Any
from unittest import mock

import botocore
import moto
import pytest
from botocore.credentials import ReadOnlyCredentials

from datacube.testutils import write_files
from datacube.utils.aws import (
    _fetch_text,
    _s3_cache_key,
    auto_find_region,
    ec2_current_region,
    get_aws_settings,
    get_creds_with_retry,
    mk_boto_session,
    obtain_new_iam_auth_token,
    s3_client,
    s3_dump,
    s3_fetch,
    s3_fmt_range,
    s3_head_object,
    s3_url_parse,
)


def _json(**kw):
    return json.dumps(kw)


def mock_urlopen(text, code=200):
    m = mock.MagicMock()
    m.getcode.return_value = code
    m.read.return_value = text.encode("utf8")
    m.__enter__.return_value = m
    return m


def test_ec2_current_region() -> None:
    tests = [
        (None, None),
        (_json(region="TT"), "TT"),
        (_json(x=3), None),
        ("not valid json", None),
    ]

    for rv, expect in tests:
        with mock.patch("datacube.utils.aws._fetch_text", return_value=rv):
            assert ec2_current_region() == expect


@mock.patch("datacube.utils.aws.botocore_default_region", return_value=None)
def test_auto_find_region(*mocks) -> None:
    with mock.patch("datacube.utils.aws._fetch_text", return_value=None):  # noqa: SIM117
        with pytest.raises(ValueError):
            auto_find_region()

    with mock.patch("datacube.utils.aws._fetch_text", return_value=_json(region="TT")):
        assert auto_find_region() == "TT"


@mock.patch(
    "datacube.utils.aws.botocore_default_region", return_value="tt-from-botocore"
)
def test_auto_find_region_2(*mocks) -> None:
    assert auto_find_region() == "tt-from-botocore"


def test_fetch_text() -> None:
    with mock.patch("datacube.utils.aws.urlopen", return_value=mock_urlopen("", 505)):
        assert _fetch_text("http://localhost:8817") is None

    with mock.patch(
        "datacube.utils.aws.urlopen", return_value=mock_urlopen("text", 200)
    ):
        assert _fetch_text("http://localhost:8817") == "text"

    def fake_urlopen(*args, **kw):
        raise OSError("Always broken")

    with mock.patch("datacube.utils.aws.urlopen", fake_urlopen):
        assert _fetch_text("http://localhost:8817") is None


def test_get_aws_settings(monkeypatch, without_aws_env) -> None:
    pp = write_files(
        {
            "config": """
[default]
region = us-west-2

[profile east]
region = us-east-1
[profile no_region]
""",
            "credentials": """
[default]
aws_access_key_id = AKIAWYXYXYXYXYXYXYXY
aws_secret_access_key = fake-fake-fake
[east]
aws_access_key_id = AKIAEYXYXYXYXYXYXYXY
aws_secret_access_key = fake-fake-fake
""",
        }
    )

    assert (pp / "credentials").exists()
    assert (pp / "config").exists()

    monkeypatch.setenv("AWS_CONFIG_FILE", str(pp / "config"))
    monkeypatch.setenv("AWS_SHARED_CREDENTIALS_FILE", str(pp / "credentials"))

    aws, creds = get_aws_settings()
    assert creds is not None
    assert aws["region_name"] == "us-west-2"
    assert aws["aws_access_key_id"] == "AKIAWYXYXYXYXYXYXYXY"
    assert aws["aws_secret_access_key"] == "fake-fake-fake"

    sess = mk_boto_session(
        profile="no_region", creds=creds.get_frozen_credentials(), region_name="mordor"
    )

    assert (
        sess.get_credentials().get_frozen_credentials()
        == creds.get_frozen_credentials()
    )

    aws, creds = get_aws_settings(profile="east")
    assert aws["region_name"] == "us-east-1"
    assert aws["aws_access_key_id"] == "AKIAEYXYXYXYXYXYXYXY"
    assert aws["aws_secret_access_key"] == "fake-fake-fake"

    aws, creds = get_aws_settings(aws_unsigned=True)
    assert creds is None
    assert aws["region_name"] == "us-west-2"
    assert aws["aws_unsigned"] is True

    aws, creds = get_aws_settings(
        profile="no_region", region_name="us-west-1", aws_unsigned=True
    )

    assert creds is None
    assert aws["region_name"] == "us-west-1"
    assert aws["aws_unsigned"] is True

    with mock.patch(
        "datacube.utils.aws._fetch_text", return_value=_json(region="mordor")
    ):
        aws, creds = get_aws_settings(profile="no_region", aws_unsigned=True)

        assert aws["region_name"] == "mordor"
        assert aws["aws_unsigned"] is True


@mock.patch("datacube.utils.aws.get_creds_with_retry", return_value=None)
def test_get_aws_settings_no_credentials(without_aws_env) -> None:
    # get_aws_settings should fail when credentials are not available
    with pytest.raises(ValueError, match="Couldn't get credentials"):
        aws, creds = get_aws_settings(region_name="fake")


def test_creds_with_retry() -> None:
    session = mock.MagicMock()
    session.get_credentials = mock.MagicMock(return_value=None)

    assert get_creds_with_retry(session, 2, 0.01) is None
    assert session.get_credentials.call_count == 2


def test_s3_url_parsing() -> None:
    assert s3_url_parse("s3://bucket/key") == ("bucket", "key")
    assert s3_url_parse("s3://bucket/key/") == ("bucket", "key/")
    assert s3_url_parse("s3://bucket/k/k/key") == ("bucket", "k/k/key")

    with pytest.raises(ValueError):
        s3_url_parse("file://some/path")


def test_s3_fmt_range() -> None:
    from numpy import s_

    assert s3_fmt_range((0, 3)) == "bytes=0-2"
    assert s3_fmt_range(s_[4:10]) == "bytes=4-9"
    assert s3_fmt_range(s_[:10]) == "bytes=0-9"
    assert s3_fmt_range(None) is None

    for bad in (s_[10:], s_[-2:3], s_[:-3], (-1, 3), (3, -1), s_[1:100:3]):
        with pytest.raises(ValueError):
            s3_fmt_range(bad)


def test_s3_client(without_aws_env) -> None:
    from botocore.credentials import ReadOnlyCredentials

    creds = ReadOnlyCredentials("fake-key", "fake-secret", None)

    # Mock AWS to make the test run faster.
    # From ~5 seconds to ~ 0.3s
    with moto.mock_aws():
        assert (
            str(s3_client(region_name="kk")._endpoint)  # type: ignore[attr-defined]
            == "s3(https://s3.kk.amazonaws.com)"
        )
        assert (
            str(s3_client(region_name="kk", use_ssl=False)._endpoint)  # type: ignore[attr-defined]
            == "s3(http://s3.kk.amazonaws.com)"
        )

        s3 = s3_client(region_name="us-west-2", creds=creds)
        assert s3 is not None


def test_s3_io(monkeypatch, without_aws_env) -> None:
    from numpy import s_

    url = "s3://bucket/file.txt"
    bucket, _ = s3_url_parse(url)
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "fake-key-id")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "fake-secret")

    with moto.mock_aws():
        s3 = s3_client(region_name="kk")
        s3.create_bucket(  # type: ignore[attr-defined]
            Bucket=bucket, CreateBucketConfiguration={"LocationConstraint": "kk"}
        )
        assert s3_dump(b"33", url, s3=s3) is True
        assert s3_fetch(url, s3=s3) == b"33"

        meta = s3_head_object(url, s3=s3)
        assert meta is not None
        assert "LastModified" in meta
        assert "ContentLength" in meta
        assert "ETag" in meta
        assert meta["ContentLength"] == 2

        assert s3_head_object(url + "-nosuch", s3=s3) is None

        assert s3_dump(b"0123456789ABCDEF", url, s3=s3) is True
        assert s3_fetch(url, range=s_[:4], s3=s3) == b"0123"
        assert s3_fetch(url, range=s_[3:8], s3=s3) == b"34567"

        with pytest.raises(ValueError):
            s3_fetch(url, range=s_[::2], s3=s3)


def test_s3_unsigned(monkeypatch, without_aws_env) -> None:
    with moto.mock_aws():
        s3 = s3_client(aws_unsigned=True)
        assert s3._request_signer.signature_version == botocore.UNSIGNED  # type: ignore[attr-defined]

        monkeypatch.setenv("AWS_UNSIGNED", "yes")
        s3 = s3_client()
        assert s3._request_signer.signature_version == botocore.UNSIGNED  # type: ignore[attr-defined]


@mock.patch("datacube.utils.aws.ec2_current_region", return_value="us-west-2")
def test_s3_client_cache(monkeypatch, without_aws_env) -> None:
    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "fake-key-id")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "fake-secret")

    # Mock AWS to make the test run faster.
    # From ~5 seconds to ~ 0.3s
    with moto.mock_aws():
        s3 = s3_client(cache=True)
        assert s3 is s3_client(cache=True)
        assert s3 is s3_client(cache="purge")
        assert s3_client(cache="purge") is None
        assert s3 is not s3_client(cache=True)

    opts: tuple[dict[str, Any], ...] = (
        {},
        {"region_name": "foo"},
        {"region_name": "bar"},
        {"profile": "foo"},
        {"profile": "foo", "region_name": "xxx"},
        {"profile": "bar"},
        {"creds": ReadOnlyCredentials("fake1", "...", None)},
        {"creds": ReadOnlyCredentials("fake1", "...", None), "region_name": "custom"},
        {"creds": ReadOnlyCredentials("fake2", "...", None)},
    )

    keys = {_s3_cache_key(**o) for o in opts}
    assert len(keys) == len(opts)


def test_obtain_new_iam_token(monkeypatch, without_aws_env) -> None:
    import moto
    from sqlalchemy.engine.url import URL

    url = URL.create(
        "postgresql",
        host="fakehost",
        database="fake_db",
        port=5432,
        username="fakeuser",
        password="definitely_a_fake_password",
    )

    monkeypatch.setenv("AWS_ACCESS_KEY_ID", "fake-key-id")
    monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "fake-secret")
    with moto.mock_aws():
        token = obtain_new_iam_auth_token(url, region_name="us-west-1")
        assert isinstance(token, str)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import pickle
from copy import deepcopy

import numpy
import pytest
from odc.geo import CRS, BoundingBox
from odc.geo.geom import polygon

from datacube.api.core import output_geobox
from datacube.model import (
    GridSpec,
    Measurement,
    MetadataType,
    Product,
    Range,
    ranges_overlap,
)
from datacube.storage import measurement_paths
from datacube.testutils import (
    mk_sample_dataset,
    mk_sample_product,
    suppress_deprecations,
)
from datacube.testutils.geom import AlbersGS
from datacube.utils.documents import InvalidDocException


def test_gridspec() -> None:
    with suppress_deprecations():
        gs = GridSpec(  # Coverage test of deprecated class
            crs=CRS("EPSG:4326"),
            tile_size=(1, 1),
            resolution=(-0.1, 0.1),
            origin=(10, 10),
        )
    poly = polygon(
        [(10, 12.2), (10.8, 13), (13, 10.8), (12.2, 10), (10, 12.2)],
        crs=CRS("EPSG:4326"),
    )
    cells = dict(list(gs.tiles_from_geopolygon(poly)))
    assert set(cells.keys()) == {(0, 1), (0, 2), (1, 0), (1, 1), (1, 2), (2, 0), (2, 1)}
    assert numpy.isclose(
        cells[(2, 0)].coordinates["longitude"].values,
        numpy.linspace(12.05, 12.95, num=10),
    ).all()
    assert numpy.isclose(
        cells[(2, 0)].coordinates["latitude"].values,
        numpy.linspace(10.95, 10.05, num=10),
    ).all()

    # check geobox_cache
    cache: dict = {}
    poly = gs.tile_geobox((3, 4)).extent
    ((c1, gbox1),) = list(gs.tiles_from_geopolygon(poly, geobox_cache=cache))
    ((c2, gbox2),) = list(gs.tiles_from_geopolygon(poly, geobox_cache=cache))

    assert c1 == (3, 4) and c2 == c1
    assert gbox1 is gbox2

    assert "4326" in str(gs)
    assert "4326" in repr(gs)
    assert gs == gs
    assert (gs == {}) is False


def test_gridspec_upperleft() -> None:
    """Test to ensure grid indexes can be counted correctly from bottom left or top left"""
    tile_bbox = BoundingBox(
        left=1934400.0,
        top=2414800.0,
        right=2084400.000,
        bottom=2264800.000,
        crs=CRS("EPSG:5070"),
    )
    bbox = BoundingBox(
        left=1934615, top=2379460, right=1937615, bottom=2376460, crs=CRS("EPSG:5070")
    )
    # Upper left - validated against WELD product tile calculator
    # http://globalmonitoring.sdstate.edu/projects/weld/tilecalc.php
    with suppress_deprecations():
        gs = GridSpec(
            crs=CRS("EPSG:5070"),
            tile_size=(-150000, 150000),
            resolution=(-30, 30),
            origin=(3314800.0, -2565600.0),
        )  # Coverage test  of deprecated class
    cells = dict(list(gs.tiles(bbox)))
    assert set(cells.keys()) == {(30, 6)}
    assert cells[(30, 6)].extent.boundingbox == tile_bbox

    with suppress_deprecations():
        gs = GridSpec(
            crs=CRS("EPSG:5070"),
            tile_size=(150000, 150000),
            resolution=(-30, 30),
            origin=(14800.0, -2565600.0),
        )
    cells = dict(list(gs.tiles(bbox)))
    assert set(cells.keys()) == {
        (30, 15)
    }  # WELD grid spec has 21 vertical cells -- 21 - 6 = 15
    assert cells[(30, 15)].extent.boundingbox == tile_bbox


def test_dataset_basics() -> None:
    ds = mk_sample_dataset([{"name": "a"}])
    assert ds == ds
    assert ds != "33"
    assert (ds == "33") is False
    assert str(ds) == repr(ds)

    ds = mk_sample_dataset([{"name": "a"}], uri=None, geobox=None)
    assert not ds.uri
    assert ds.uri_scheme == ""
    assert ds.crs is None
    assert ds.bounds is None
    assert ds.extent is None
    assert ds.transform is None


def test_dataset_measurement_paths() -> None:
    format_ = "GeoTiff"

    ds = mk_sample_dataset(
        [{"name": n, "path": n + ".tiff"} for n in ["a", "b", "c"]],
        uri="file:///tmp/datataset.yml",
        format=format_,
    )

    assert ds.local_uri == ds.uri
    assert ds.legacy_uri() == ds.uri
    assert ds.uri_scheme == "file"
    assert ds.format == format_
    paths = measurement_paths(ds)

    for k, v in paths.items():
        assert v == "file:///tmp/" + k + ".tiff"

    ds._uris = []
    ds.uri = None
    assert ds.local_uri is None
    with pytest.raises(ValueError):
        measurement_paths(ds)


def test_product_basics() -> None:
    product = mk_sample_product("test_product")
    assert product.name == "test_product"
    assert "test_product" in str(product)
    assert "test_product" in repr(product)
    assert product == product
    assert product == mk_sample_product("test_product")
    assert product != mk_sample_product("other")
    assert product != [()]
    assert hash(product) == hash(mk_sample_product("test_product"))
    assert "time" in dir(product.metadata)

    assert product.measurements == product.lookup_measurements()
    assert product.lookup_measurements(["red"]) == product.lookup_measurements("red")


def test_product_dimensions() -> None:
    product = mk_sample_product("test_product")
    with suppress_deprecations():
        assert product.grid_spec is None
    assert product.dimensions == ("time", "y", "x")

    product = mk_sample_product("test_product", with_grid_spec=True)
    with suppress_deprecations():
        assert product.grid_spec is not None
    assert product.dimensions == ("time", "y", "x")

    partial_storage = product.definition["storage"]
    partial_storage.pop("tile_size")
    product = mk_sample_product("tt", storage=partial_storage)
    with suppress_deprecations():
        assert product.grid_spec is None


def test_product_gridspec() -> None:
    product = mk_sample_product("old_product", with_grid_spec=True)
    with pytest.deprecated_call():
        assert product.grid_spec is not None
    assert product.to_dict()["tile_shape"] == (4000.0, 4000.0)

    storage = {
        "crs": "EPSG:3577",
        "resolution": 25,
        "tile_shape": {"x": 4000.0, "y": 4000.0},
    }
    product = mk_sample_product("new_product", storage=storage)
    assert product.grid_spec is not None
    assert not isinstance(product.grid_spec, GridSpec)
    assert product.to_dict()["resolution"].xy == (25, -25)


def test_product_nodata_nan() -> None:
    # When storing .nan to JSON in DB it becomes a string with value "NaN"
    # Make sure it is converted back to real NaN
    product = mk_sample_product(
        "test",
        measurements=[
            {"name": "_nan", "dtype": "float32", "nodata": "NaN"},
            {"name": "_inf", "dtype": "float32", "nodata": "Infinity"},
            {"name": "_neg_inf", "dtype": "float32", "nodata": "-Infinity"},
        ],
    )
    for m in product.measurements.values():
        assert isinstance(m.nodata, float)

    assert numpy.isnan(product.measurements["_nan"].nodata)
    assert product.measurements["_inf"].nodata == numpy.inf
    assert product.measurements["_neg_inf"].nodata == -numpy.inf


def test_product_scale_factor() -> None:
    product = mk_sample_product(
        "test", measurements=[{"name": "red", "scale_factor": 33, "add_offset": -5}]
    )
    assert product.validate(product.definition) is None  # type: ignore[attr-defined]
    assert product.measurements["red"].scale_factor == 33
    assert product.measurements["red"].add_offset == -5
    attrs = product.measurements["red"].dataarray_attrs()
    assert attrs["scale_factor"] == 33
    assert attrs["add_offset"] == -5


def test_product_load_hints() -> None:
    product = mk_sample_product(
        "test_product", load={"crs": "epsg:3857", "resolution": {"x": 10, "y": -10}}
    )

    assert "load" in product.definition
    assert Product.validate(product.definition) is None  # type: ignore[attr-defined]

    hints = product._extract_load_hints()
    assert hints is not None
    assert hints["crs"] == CRS("epsg:3857")
    assert hints["resolution"].yx == (-10, 10)
    assert "align" not in hints

    product = mk_sample_product(
        "test_product",
        load={
            "crs": "epsg:3857",
            "align": {"x": 5, "y": 6},
            "resolution": {"x": 10, "y": -10},
        },
    )

    hints = product.load_hints()
    assert hints["output_crs"] == CRS("epsg:3857")
    assert hints["resolution"].yx == (-10, 10)
    assert hints["align"] == (6, 5)
    assert product.default_crs == CRS("epsg:3857")
    default_resolution = product.default_resolution
    assert default_resolution is not None
    assert default_resolution.yx == (-10, 10)
    assert product.default_align == (6, 5)

    product = mk_sample_product(
        "test_product",
        load={
            "crs": "epsg:4326",
            "align": {"longitude": 0.5, "latitude": 0.6},
            "resolution": {"longitude": 1.2, "latitude": -1.1},
        },
    )

    hints = product.load_hints()
    assert hints["output_crs"] == CRS("epsg:4326")
    assert hints["resolution"].yx == (-1.1, 1.2)
    assert hints["align"] == (0.6, 0.5)

    # check it's cached
    assert product.load_hints() is product.load_hints()

    # check schema: crs and resolution are compulsory
    for k in ("resolution", "crs"):
        doc = deepcopy(product.definition)
        assert Product.validate(doc) is None  # type: ignore[attr-defined]

        doc["load"].pop(k)
        assert k not in doc["load"]

        with pytest.raises(InvalidDocException):
            Product.validate(doc)  # type: ignore[attr-defined]

    # check GridSpec leakage doesn't happen for fully defined gridspec
    product = mk_sample_product("test", with_grid_spec=True)
    with suppress_deprecations():
        assert product.grid_spec is not None
    assert product.load_hints() == {}

    # check for fallback into partially defined `storage:`
    product = mk_sample_product(
        "test", storage={"crs": "EPSG:3857", "resolution": {"x": 10, "y": -10}}
    )
    with suppress_deprecations():
        assert product.grid_spec is None
    default_resolution = product.default_resolution
    assert default_resolution is not None
    assert default_resolution.yx == (-10, 10)
    assert product.default_crs == CRS("EPSG:3857")

    # check for fallback into partially defined `storage:`
    # no resolution -- no hints
    product = mk_sample_product("test", storage={"crs": "EPSG:3857"})
    with suppress_deprecations():
        assert product.grid_spec is None
    assert product.load_hints() == {}

    # check misspelled load hints
    product = mk_sample_product(
        "test_product",
        load={"crs": "epsg:4326", "resolution": {"longtude": 1.2, "latitude": -1.1}},
    )
    assert product.load_hints() == {}


def test_measurement() -> None:
    # Can create a measurement
    m = Measurement(name="t", dtype="uint8", nodata=255, units="1")

    # retrieve it's vital stats
    assert m.name == "t"
    assert m.dtype == "uint8"
    assert m.nodata == 255
    assert m.units == "1"

    # retrieve the information required for filling a DataArray
    assert m.dataarray_attrs() == {"nodata": 255, "units": "1"}

    # Can add a new attribute by name and ensure it updates the DataArray attrs too
    m["bob"] = 10
    assert m.bob == 10
    assert m.dataarray_attrs() == {"nodata": 255, "units": "1", "bob": 10}

    m["none"] = None
    assert m.none is None

    # Resampling method is special and *not* needed for DataArray attrs
    m["resampling_method"] = "cubic"
    assert "resampling_method" not in m.dataarray_attrs()

    # It's possible to copy and update a Measurement instance
    m2 = m.copy()
    assert m2.bob == 10
    assert m2.dataarray_attrs() == m.dataarray_attrs()

    assert repr(m2) == repr(m)

    # Must specify *all* required keys. name, dtype, nodata and units
    with pytest.raises(ValueError) as e:
        Measurement(name="x", units="1", nodata=0)

    assert "required keys missing:" in str(e.value)
    assert "dtype" in str(e.value)


def test_measurement_equality() -> None:
    m1 = Measurement(name="t", dtype="uint8", nodata=255, units="1")
    m2 = Measurement(name="t", dtype="uint8", nodata=255, units="1")
    assert m1 == m2
    g1 = AlbersGS.tile_geobox((15, -40))
    assert m1 != g1


@pytest.mark.parametrize("protocol", range(pickle.HIGHEST_PROTOCOL))
def test_measurement_pickling(protocol) -> None:
    m = Measurement(canonical_name="s", name="t", dtype="uint8", nodata=255, units="1")

    serialised_m = pickle.dumps(m, protocol=protocol)

    restored_m = pickle.loads(serialised_m)

    assert m == restored_m


def test_measurement_cloudpickle() -> None:
    import cloudpickle

    m = Measurement(canonical_name="s", name="t", dtype="uint8", nodata=255, units="1")
    serialised = cloudpickle.dumps(m)

    deserialised = cloudpickle.loads(serialised)
    assert m == deserialised


def test_output_geobox_load_hints() -> None:
    geobox0 = AlbersGS.tile_geobox((15, -40))

    geobox = output_geobox(
        load_hints={"output_crs": geobox0.crs, "resolution": geobox0.resolution},
        geopolygon=geobox0.extent,
    )
    assert geobox == geobox0


def test_like_geobox() -> None:
    geobox = AlbersGS.tile_geobox((15, -40))
    assert output_geobox(like=geobox) is geobox


def test_output_geobox_fail_paths() -> None:
    with pytest.raises(ValueError):
        output_geobox()

    with pytest.raises(ValueError):
        output_geobox(output_crs="EPSG:4326")  # need resolution as well

    # need bounds
    with pytest.raises(ValueError):
        output_geobox(output_crs="EPSG:4326", resolution=(1, 1))


def test_metadata_type() -> None:
    m = MetadataType(
        {
            "name": "eo",
            "dataset": {
                "id": ["id"],
                "label": ["ga_label"],
                "creation_time": ["creation_dt"],
                "measurements": ["image", "bands"],
                "sources": ["lineage", "source_datasets"],
                "format": ["format", "name"],
            },
        },
    )

    assert "eo" in str(m)
    assert "eo" in repr(m)
    assert m.name == "eo"
    assert m.description is None
    assert m.dataset_reader({}) is not None


def test_ranges_overlap() -> None:
    assert not ranges_overlap(Range(begin=1, end=5), Range(begin=11, end=15))
    assert not ranges_overlap(Range(begin=1, end=5), Range(begin=5, end=11))
    assert not ranges_overlap(Range(begin=5, end=11), Range(begin=1, end=5))
    assert ranges_overlap(Range(begin=1, end=5), Range(begin=1, end=5))
    assert ranges_overlap(Range(begin=1, end=15), Range(begin=10, end=25))
    assert ranges_overlap(Range(begin=10, end=25), Range(begin=1, end=15))
    assert ranges_overlap(Range(begin=1, end=25), Range(begin=10, end=15))
    assert ranges_overlap(Range(begin=10, end=15), Range(begin=1, end=25))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

from ..index.abstract import AbstractIndexDriver
from ._tools import singleton_setup
from .driver_cache import load_drivers


class IndexDriverCache:
    def __init__(self, group: str) -> None:
        self._drivers = load_drivers(group)

        if len(self._drivers) == 0:
            # Driver load has failed.  Something is very wrong, but try to manually load the main drivers anyway.
            from datacube.index.memory.index import (
                index_driver_init as mem_index_driver_init,
            )
            from datacube.index.postgis.index import (
                index_driver_init as pgis_index_driver_init,
            )
            from datacube.index.postgres.index import (
                index_driver_init as pg_index_driver_init,
            )

            self._drivers = {
                "postgres": pg_index_driver_init(),
                "postgis": pgis_index_driver_init(),
                "memory": mem_index_driver_init(),
            }

        for driver in list(self._drivers.values()):
            if hasattr(driver, "aliases"):
                for alias in driver.aliases:
                    self._drivers[alias] = driver

    def __call__(self, name: str) -> AbstractIndexDriver | None:
        """
        :returns: None if driver with a given name is not found

        :param name: Driver name
        :return: Returns IndexDriver
        """
        return self._drivers.get(name, None)

    def drivers(self) -> list[str]:
        """
        Returns list of driver names
        """
        return list(self._drivers.keys())


def index_cache() -> IndexDriverCache:
    """
    Singleton for IndexDriverCache
    """
    return singleton_setup(
        index_cache, "_instance", IndexDriverCache, "datacube.plugins.index"
    )


def index_drivers() -> set[str]:
    """
    Returns a set of driver names
    """
    return set(index_cache().drivers())


def index_driver_by_name(name: str) -> AbstractIndexDriver | None:
    """
    Lookup index driver by name

    :returns: Initialised index driver instance
    :returns: None if driver with this name doesn't exist
    """
    return index_cache()(name)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock

import pytest

from datacube.testutils import suppress_deprecations


def test_smells_like_ini() -> None:
    from datacube.cfg.utils import smells_like_ini

    assert smells_like_ini("""[an_ini_file]
key: value
other-key: 12

""")
    assert smells_like_ini("""
; This is an ini comment
[an_ini_file]
key: value
other-key: 12

""")
    assert not smells_like_ini("""# This is a YAML comment

spagoots:
  four: score
  and: 7
  years: ago
""")
    assert not smells_like_ini("""{
"spagoots":{
  "four": "score",
  "and": 7,
  "years": "ago",
  # no comments in json!
  }
}
""")
    assert smells_like_ini("""[
    Looks ini-ish but
    :::is:::
    GIBBER-ish ..;';'@#$%^@#$%^#$^&$
\n  \n    \n \n  \t  \t  \n   \n
aasdfer\\faw043[]]][""")

    # Pure white space
    assert smells_like_ini("   \n  \n    \n \n  \t  \t  \n   \n")


@pytest.fixture
def simple_valid_ini() -> str:
    return """[foo]
bar: bell
bat: 7
baz: luhrmann
[goo]
bar: beque
bat: 2
baz: mckenzie
"""


@pytest.fixture
def simple_valid_yaml() -> str:
    return """foo:
    bar: bell
    bat: 7
    baz: luhrmann
goo:
    bar: beque
    bat: 2
    baz: mckenzie
"""


def test_parse_text(simple_valid_ini, simple_valid_yaml) -> None:
    from datacube.cfg import CfgFormat, ConfigException, parse_text

    ini = parse_text(simple_valid_ini)
    yaml = parse_text(simple_valid_yaml)
    assert ini["foo"]["bar"] == yaml["foo"]["bar"]
    assert ini["goo"]["baz"] == yaml["goo"]["baz"]
    assert int(ini["foo"]["bat"]) == int(yaml["foo"]["bat"])

    with pytest.raises(ConfigException):
        parse_text(simple_valid_ini, fmt=CfgFormat.YAML)

    with pytest.raises(ConfigException):
        parse_text(simple_valid_yaml, fmt=CfgFormat.INI)


def mk_single_env_config(connector: str) -> str:
    return f"""# Simple single environment config
new:
   index_driver: postgis
   db_url: postgresql{connector}://foo:bar@server.subdomain.domain/mytestdb
   db_iam_authentication: yes
"""


@pytest.fixture
def single_env_config():
    return mk_single_env_config("+psycopg2")


@pytest.fixture
def single_env_config_no_connector():
    return mk_single_env_config("")


@pytest.fixture
def simple_config() -> str:
    return """# Simple but thorough test config
default:
   alias: legacy
postgres:
   alias: legacy
legacy:
   index_driver: default
   db_username: foo
   db_password: bar
   db_hostname: server.subdomain.domain
   db_port: 5433
   db_database: mytestdb
   db_connection_timeout: 20
new:
   index_driver: postgis
   db_url: postgresql+psycopg2://foo:bar@server.subdomain.domain/mytestdb
   db_iam_authentication: yes
postgis:
   alias: new
memory:
   index_driver: memory
   db_url: '@nota?valid:url//foo&bar%%%'
new2:
   index_driver: postgis
   db_url: postgresql+psycopg2://foo:bar@server.subdomain.domain/mytestdb
   db_database: not_read
   db_port: ignored
   db_iam_authentication: yes
   db_iam_timeout: 300
"""


@pytest.fixture
def simple_dict():
    return {
        "default": {"alias": "legacy"},
        "postgres": {"alias": "legacy"},
        "legacy": {
            "index_driver": "default",
            "db_username": "foo",
            "db_password": "bar",
            "db_hostname": "server.subdomain.domain",
            "db_port": 5433,
            "db_database": "mytestdb",
            "db_connection_timeout": 20,
        },
        "new": {
            "index_driver": "postgis",
            "db_url": "postgresql+psycopg2://foo:bar@server.subdomain.domain/mytestdb",
            "db_iam_authentication": "yes",
        },
        "postgis": {"alias": "new"},
        "memory": {"index_driver": "memory", "db_url": "@nota?valid:url//foo&bar%%%"},
        "new2": {
            "index_driver": "postgis",
            "db_url": "postgresql+psycopg2://foo:bar@server.subdomain.domain/mytestdb",
            "db_database": "not_read",
            "db_port": "ignored",
            "db_iam_authentication": "yes",
            "db_iam_timeout": 300,
        },
    }


def test_invalid_env() -> None:
    from datacube.cfg import ConfigException, ODCConfig

    with pytest.raises(ConfigException):
        ODCConfig(
            text="""
default:
    alias: royale_avec_fromage
            """
        )
    with pytest.raises(ConfigException):
        ODCConfig(
            text="""
default:
    alias: legit00
non_legit:
    index_driver: null
        """
        )


def test_oldstyle_cfg() -> None:
    from datacube.cfg import ODCConfig

    with pytest.warns(UserWarning, match=r"default_environment.*no longer supported"):
        cfg = ODCConfig(
            text="""
        default:
            index_driver: memory
        env2:
            index_driver: memory
        env3:
            index_driver: memory
        user:
            default_environment: env3
        """
        )
        assert cfg[None]._name == "default"


def test_invalid_option() -> None:
    from datacube.cfg import ConfigException, ODCOptionHandler

    mockenv = MagicMock()
    with pytest.raises(ConfigException):
        ODCOptionHandler("NO_CAPS", mockenv)


def test_single_env(single_env_config, single_env_config_no_connector) -> None:
    from datacube.cfg import ODCConfig

    db_urls = []
    for cfg in [
        ODCConfig(text=single_env_config),
        ODCConfig(text=single_env_config_no_connector),
    ]:
        assert cfg["new"].index_driver == "postgis"
        db_urls.append(cfg["new"].db_url)
        assert cfg["new"].db_username == "foo"
        with pytest.raises(AttributeError):
            assert cfg["new"].not_an_option
        assert cfg["new"]["db_iam_authentication"]
        assert cfg["new"].db_iam_timeout == 600
        assert cfg["new"]["db_connection_timeout"] == 60
    assert db_urls == [
        "postgresql+psycopg2://foo:bar@server.subdomain.domain/mytestdb",
        "postgresql://foo:bar@server.subdomain.domain/mytestdb",
    ]


def assert_simple_aliases(cfg) -> None:
    assert cfg["default"]._name == "legacy"
    assert cfg["postgres"]._name == "legacy"
    assert cfg["legacy"]._name == "legacy"
    assert cfg["postgis"]._name == "new"
    assert cfg["new"]._name == "new"
    assert cfg["memory"]._name == "memory"
    assert cfg["new2"]._name == "new2"
    assert cfg["dynamic"]._name == "dynamic"
    assert cfg[None]._name == "legacy"


def test_aliases(simple_config) -> None:
    from datacube.cfg import ConfigException, ODCConfig

    cfg = ODCConfig(text=simple_config)
    assert_simple_aliases(cfg)
    with pytest.raises(ConfigException) as e:
        cfg = ODCConfig(
            raw_dict={
                "default": {"alias": "main", "invalid_option": True},
                "main": {"index_driver": "memory"},
            }
        )
    assert "invalid_option" in str(e.value)


def assert_simple_options(cfg) -> None:
    assert cfg["default"]["index_driver"] == "default"
    assert cfg["default"].db_username == "foo"
    assert not cfg["default"]["db_iam_authentication"]
    with pytest.raises(KeyError):
        assert cfg["default"]["db_iam_timeout"]

    assert (
        cfg["new2"].db_url
        == "postgresql+psycopg2://foo:bar@server.subdomain.domain/mytestdb"
    )
    assert cfg["new2"].db_username == "foo"
    with pytest.raises(AttributeError):
        assert cfg["new2"].not_an_option
    assert cfg["new2"]["db_iam_authentication"]
    assert cfg["new2"].db_iam_timeout == 300
    assert cfg["new2"]["db_connection_timeout"] == 60


def test_options(simple_config) -> None:
    from datacube.cfg import ODCConfig

    cfg = ODCConfig(text=simple_config)
    assert_simple_options(cfg)


def test_rawdict(simple_dict) -> None:
    from datacube.cfg import ODCConfig

    cfg = ODCConfig(raw_dict=simple_dict)
    assert_simple_aliases(cfg)
    assert_simple_options(cfg)


def test_noenv_overrides_in_text(simple_config, monkeypatch) -> None:
    monkeypatch.setenv("ODC_LEGACY_DB_USERNAME", "bar")
    monkeypatch.setenv("ODC_NEW_DB_USERNAME", "bar")
    from datacube.cfg import ODCConfig

    cfg = ODCConfig(text=simple_config)

    assert cfg["legacy"].db_username != "bar"
    assert cfg["new"].db_username != "bar"


@pytest.fixture
def path_to_yaml_config():
    import os.path

    return os.path.join(os.path.dirname(__file__), "cfg", "simple_cfg.yaml")


@pytest.fixture
def path_to_ini_config():
    import os.path

    return os.path.join(os.path.dirname(__file__), "cfg", "simple_cfg.ini")


@pytest.fixture
def path_to_different_config():
    import os.path

    return os.path.join(os.path.dirname(__file__), "cfg", "different_cfg.yaml")


def test_yaml_from_path(path_to_yaml_config) -> None:
    from datacube.cfg import ODCConfig

    cfg = ODCConfig(paths=path_to_yaml_config)
    assert_simple_aliases(cfg)
    assert_simple_options(cfg)


def test_ini_from_path(path_to_ini_config) -> None:
    from datacube.cfg import ODCConfig

    cfg = ODCConfig(paths=path_to_ini_config)
    assert_simple_aliases(cfg)
    assert_simple_options(cfg)


def test_ini_from_paths(
    path_to_ini_config, path_to_yaml_config, path_to_different_config, monkeypatch
) -> None:
    from datacube.cfg import ConfigException, ODCConfig

    cfg = ODCConfig(
        paths=[
            "/non/existent/path.yml",
            path_to_yaml_config,
            path_to_different_config,
            path_to_ini_config,
        ]
    )
    assert cfg[None]._name == "legacy"

    cfg = ODCConfig(
        paths=[
            path_to_different_config,
            "/non/existent/path.yml",
            path_to_yaml_config,
        ]
    )
    assert cfg[None]._name == "new"

    with pytest.raises(ConfigException):
        cfg = ODCConfig(
            paths=[
                "/non/existent/path.yml",
                "/another/nonexistent/path.yml",
                "/etc",
            ]
        )

    with monkeypatch.context() as mp:
        mp.setattr(
            "datacube.cfg.cfg._DEFAULT_CONFIG_SEARCH_PATH",
            [
                "/non/existent/path.yml",
                "/another/nonexistent/path.yml",
                "/etc",
                path_to_yaml_config,
            ],
        )
        cfg = ODCConfig()
        assert cfg[None]._name == "legacy"

    with monkeypatch.context() as mp:
        mp.setattr(
            "datacube.cfg.cfg._DEFAULT_CONFIG_SEARCH_PATH",
            [
                "/non/existent/path.yml",
                "/etc",
                "/another/nonexistent/path.yml",
            ],
        )
        cfg = ODCConfig()
        assert cfg[None]._name == "default"

    with monkeypatch.context() as mp:
        mp.setenv(
            "ODC_CONFIG_PATH",
            f"/non/existent/path.yml:{path_to_yaml_config}:{path_to_different_config}",
        )
        cfg = ODCConfig()
        assert cfg[None]._name == "legacy"

    with suppress_deprecations():  # noqa: SIM117
        with monkeypatch.context() as mp:
            mp.setenv(
                "DATACUBE_CONFIG_PATH",
                f"/non/existent/path.yml:{path_to_yaml_config}:{path_to_different_config}",
            )
            cfg = ODCConfig()
            assert cfg[None]._name == "legacy"


def test_envvar_overrides(path_to_yaml_config, monkeypatch) -> None:
    monkeypatch.setenv("ODC_LEGACY_DB_USERNAME", "bar")
    monkeypatch.setenv("ODC_NEW_DB_USERNAME", "bar")
    monkeypatch.setenv("ODC_NEW2_DB_CONNECTION_TIMEOUT", "20")
    monkeypatch.setenv("DATACUBE_IAM_AUTHENTICATION", "yes")

    from datacube.cfg import ODCConfig

    with suppress_deprecations():
        cfg = ODCConfig(paths=path_to_yaml_config)
        assert cfg["legacy"].db_username == "bar"
        assert cfg["legacy"].db_iam_authentication
        assert cfg["new"].db_iam_authentication
        assert cfg["new2"].db_iam_authentication
        assert cfg["new2"].db_connection_timeout == 20
        assert cfg["new"].db_username != "bar"


def test_intopt_validation() -> None:
    from datacube.cfg import ConfigException, ODCConfig

    cfg = ODCConfig(
        text="""
env1:
    db_hostname: localhost
    db_port: seven
"""
    )
    with pytest.raises(ConfigException):
        assert cfg["env1"].db_hostname == "localhost"
    cfg = ODCConfig(
        text="""
env1:
    db_hostname: localhost
    db_port: -7
"""
    )
    with pytest.raises(ConfigException):
        assert cfg["env1"].db_hostname == "localhost"
    cfg = ODCConfig(
        text="""
env1:
    db_hostname: localhost
    db_port: 4425542239934
"""
    )
    with pytest.raises(ConfigException):
        assert cfg["env1"].db_hostname == "localhost"
    cfg = ODCConfig(
        text="""
env1:
    db_hostname: localhost
    db_port: 0
"""
    )
    with pytest.raises(ConfigException):
        assert cfg["env1"].db_hostname == "localhost"


def test_invalid_idx_driver() -> None:
    from datacube.cfg import ConfigException, ODCConfig

    cfg = ODCConfig(
        raw_dict={
            "default": {"alias": "foo"},
            "foo": {
                "index_driver": "phillips_head",
            },
        }
    )
    with pytest.raises(ConfigException) as e:
        _ = cfg["default"].index_driver
    estr = str(e.value)
    assert "Unknown index driver" in estr
    assert "phillips_head" in estr
    assert "legacy" in estr


def test_invalid_pg_url() -> None:
    from datacube.cfg import ConfigException, ODCConfig

    cfg = ODCConfig(
        raw_dict={
            "default": {"alias": "foo"},
            "foo": {"index_driver": "postgres", "db_url": "https://www.google.com"},
        }
    )
    with pytest.raises(ConfigException):
        assert cfg["default"].index_driver == "postgres"


def test_pgurl_from_config(simple_dict) -> None:
    from datacube.cfg import ODCConfig, psql_url_from_config

    cfg = ODCConfig(raw_dict=simple_dict)
    assert (
        psql_url_from_config(cfg["legacy"])
        == "postgresql+psycopg2://foo:bar@server.subdomain.domain:5433/mytestdb"
    )
    assert (
        psql_url_from_config(cfg["new"])
        == "postgresql+psycopg2://foo:bar@server.subdomain.domain/mytestdb"
    )
    with pytest.raises(AttributeError):
        psql_url_from_config(cfg["memory"])
    assert (
        cfg["new2"].db_url
        == "postgresql+psycopg2://foo:bar@server.subdomain.domain/mytestdb"
    )
    assert cfg["new2"].db_username == "foo"
    assert cfg["new2"].db_password == "bar"
    assert cfg["new2"].db_hostname == "server.subdomain.domain"
    assert not cfg["new2"].db_port
    assert cfg["new2"].db_database == "mytestdb"

    cfg = ODCConfig(
        raw_dict={
            "foo": {
                "db_hostname": "remotehost.local",
                "db_username": "penelope",
                "db_database": "mydb",
                "db_port": 5544,
            }
        }
    )
    assert (
        psql_url_from_config(cfg["foo"])
        == "postgresql+psycopg2://penelope@remotehost.local:5544/mydb"
    )
    cfg = ODCConfig(
        raw_dict={
            "foo": {
                "db_hostname": "remotehost.local",
                "db_username": "penelope",
                "db_port": 5544,
            }
        }
    )


def test_multiple_sourcetypes(simple_config, path_to_ini_config, simple_dict) -> None:
    from datacube.cfg import ConfigException, ODCConfig

    with pytest.raises(ConfigException) as e:
        ODCConfig(paths=path_to_ini_config, raw_dict=simple_dict, text=simple_config)
    assert "Can only supply one of" in str(e.value)
    with pytest.raises(ConfigException) as e:
        ODCConfig(raw_dict=simple_dict, text=simple_config)
    assert "Can only supply one of" in str(e.value)
    with pytest.raises(ConfigException) as e:
        ODCConfig(paths=path_to_ini_config, text=simple_config)
    assert "Can only supply one of" in str(e.value)
    with pytest.raises(ConfigException) as e:
        ODCConfig(paths=path_to_ini_config, raw_dict=simple_dict)
    assert "Can only supply one of" in str(e.value)


def test_get_environment(simple_config) -> None:
    from datacube.cfg import ConfigException, ODCConfig

    cfg = ODCConfig(text=simple_config)
    with pytest.raises(ConfigException) as e:
        ODCConfig.get_environment(config=cfg, raw_config=simple_config)
    assert "Cannot specify both" in str(e.value)
    env = ODCConfig.get_environment(config=cfg, env="default")
    assert env is cfg[None]


def test_raw_by_environment(simple_config, monkeypatch) -> None:
    from datacube.cfg import ODCConfig

    monkeypatch.setenv(
        "ODC_CONFIG",
        '{"default":{"alias": "foo"},"foo":{"index_driver":"postgis","db_url":"postgresql+psycopg2:///mydb"}}',
    )
    cfg = ODCConfig()
    assert cfg[None]._name == "foo"


def test_default_environment(simple_config, monkeypatch) -> None:
    from datacube.cfg import ODCConfig

    with suppress_deprecations():
        cfg = ODCConfig(text=simple_config)
        assert cfg[None]._name == "legacy"
        monkeypatch.setenv("ODC_ENVIRONMENT", "new2")
        assert cfg[None]._name == "new2"
        monkeypatch.setenv("ODC_ENVIRONMENT", "")
        monkeypatch.setenv("DATACUBE_ENVIRONMENT", "postgis")
        assert cfg[None]._name == "new"
        cfg = ODCConfig(raw_dict={"datacube": {"index_driver": "memory"}})
        monkeypatch.setenv("DATACUBE_ENVIRONMENT", "")
        assert cfg[None]._name == "datacube"
        cfg = ODCConfig(
            raw_dict={
                "weirdname": {"index_driver": "memory"},
                "stupidenv": {"index_driver": "null"},
            }
        )
        assert cfg[None]._name == "default"



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Module
"""

from collections.abc import Mapping, Sequence
from pathlib import Path

import pytest

from datacube.testutils import assert_file_structure, write_files
from datacube.ui.common import (
    _find_any_metadata_suffix,
    get_metadata_path,
    ui_path_doc_stream,
)


def test_get_metadata_path() -> None:
    test_file_structure: Mapping[
        str, str | Sequence[str] | Mapping[str, str | Sequence[str]]
    ] = {
        "directory_dataset": {
            "file1.txt": "",
            "file2.txt": "",
            "agdc-metadata.yaml.gz": "",
        },
        "file_dataset.tif": "",
        "file_dataset.tif.agdc-md.yaml": "",
        "dataset_metadata.yaml": "",
        "no_metadata.tif": "",
    }

    out_dir = write_files(test_file_structure)

    assert_file_structure(out_dir, test_file_structure)

    # A metadata file can be specified directly.
    path = get_metadata_path(out_dir.joinpath("dataset_metadata.yaml"))
    assert Path(path).absolute() == out_dir.joinpath("dataset_metadata.yaml").absolute()

    # A dataset directory will have an internal 'agdc-metadata' file.
    path = get_metadata_path(out_dir.joinpath("directory_dataset"))
    assert (
        Path(path).absolute()
        == out_dir.joinpath("directory_dataset", "agdc-metadata.yaml.gz").absolute()
    )

    # Other out_dir can have a sibling file ending in 'agdc-md.yaml'
    path = get_metadata_path(out_dir.joinpath("file_dataset.tif"))
    assert (
        Path(path).absolute()
        == out_dir.joinpath("file_dataset.tif.agdc-md.yaml").absolute()
    )

    # URLs are always themselves
    example_url = "http://localhost/dataset.yaml"
    url = get_metadata_path(example_url)
    assert url == example_url

    # Lack of metadata raises an error.
    with pytest.raises(ValueError):
        get_metadata_path(out_dir.joinpath("no_metadata.tif"))

    # Nonexistent dataset raises a ValueError.
    with pytest.raises(ValueError):
        get_metadata_path(out_dir.joinpath("missing-dataset.tif"))


def test_find_any_metatadata_suffix() -> None:
    files = write_files(
        {
            "directory_dataset": {
                "file1.txt": "",
                "file2.txt": "",
                "agdc-metadata.json.gz": "",
            },
            "file_dataset.tif.agdc-md.yaml": "",
            "dataset_metadata.YAML": "",
            "no_metadata.tif": "",
            "ambigous.yml": "",
            "ambigous.yaml": "",
        }
    )

    path = _find_any_metadata_suffix(files.joinpath("dataset_metadata"))
    assert path is not None
    assert Path(path).absolute() == files.joinpath("dataset_metadata.YAML").absolute()

    path = _find_any_metadata_suffix(
        files.joinpath("directory_dataset", "agdc-metadata")
    )
    assert path is not None
    assert (
        Path(path).absolute()
        == files.joinpath("directory_dataset", "agdc-metadata.json.gz").absolute()
    )

    path = _find_any_metadata_suffix(files.joinpath("file_dataset.tif.agdc-md"))
    assert path is not None
    assert (
        Path(path).absolute()
        == files.joinpath("file_dataset.tif.agdc-md.yaml").absolute()
    )

    # Returns none if none exist
    path = _find_any_metadata_suffix(files.joinpath("no_metadata"))
    assert path is None

    with pytest.raises(ValueError):
        _find_any_metadata_suffix(files.joinpath("ambigous"))


def test_ui_path_doc_stream(httpserver) -> None:
    filename = "dataset_metadata.yaml"
    file_content = ""
    out_dir = write_files({filename: file_content})

    httpserver.expect_request(filename).respond_with_data(file_content)

    input_paths = [
        Path(out_dir) / "dataset_metadata.yaml",
        httpserver.url_for(filename),
    ]

    for input_path, (doc, resolved_path) in zip(
        input_paths, ui_path_doc_stream(input_paths)
    ):
        assert doc == {}
        assert input_path == resolved_path



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Config reading/loading exceptions
"""


class ConfigException(Exception):  # noqa: N818
    """
    Exception raised in the configuration API
    """



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import logging
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable, Mapping, Sequence
from datetime import timedelta
from time import monotonic
from typing import Any, NamedTuple
from uuid import UUID

from deprecat import deprecat
from odc.geo import CRS, Geometry

from datacube.migration import ODC2DeprecationWarning
from datacube.model import Dataset, Field, Product, QueryDict, QueryField, Range
from datacube.utils import report_to_user
from datacube.utils.changes import AllowPolicy, Change, DocumentMismatchError, Offset
from datacube.utils.documents import JsonDict

from ._types import DSID, BatchStatus, DatasetTuple

_LOG: logging.Logger = logging.getLogger(__name__)


class AbstractDatasetResource(ABC):
    """
    Abstract base class for the Dataset portion of an index api.

    All DatasetResource implementations should inherit from this base
    class and implement all abstract methods.

    (If a particular abstract method is not applicable for a particular implementation
    raise a NotImplementedError)
    """

    def __init__(self, index) -> None:
        self._index = index
        self.products = self._index.products
        self.types = self.products  # types is compatibility alias for products

    @abstractmethod
    def get_unsafe(
        self,
        id_: DSID,
        include_sources: bool = False,
        include_deriveds: bool = False,
        max_depth: int = 0,
    ) -> Dataset:
        """
        Get dataset by id (Raises KeyError if id\\_ does not exist)

        - Index drivers supporting the legacy lineage API:

        :param id\\_: id of the dataset to retrieve
        :param include_sources: include the full provenance tree of the dataset.


        - Index drivers supporting the external lineage API:

        :param id\\_: id of the dataset to retrieve
        :param include_sources: include the full provenance tree for the dataset.
        :param include_deriveds: include the full derivative tree for the dataset.
        :param max_depth: The maximum depth of the source and/or derived tree.  Defaults to 0, meaning no limit.
        """

    def get(
        self,
        id_: DSID,
        include_sources: bool = False,
        include_deriveds: bool = False,
        max_depth: int = 0,
    ) -> Dataset | None:
        """
        Get dataset by id (Return None if ``id_`` does not exist).

        - Index drivers supporting the legacy lineage API:

        :param id\\_: id of the dataset to retrieve
        :param include_sources: include the full provenance tree of the dataset.


        - Index drivers supporting the external lineage API:

        :param id\\_: id of the dataset to retrieve
        :param include_sources: include the full provenance tree for the dataset.
        :param include_deriveds: include the full derivative tree for the dataset.
        :param max_depth: The maximum depth of the source and/or derived tree.  Defaults to 0, meaning no limit.
        """
        try:
            return self.get_unsafe(id_, include_sources, include_deriveds, max_depth)
        except KeyError:
            return None

    def _check_get_legacy(
        self, include_deriveds: bool = False, max_depth: int = 0
    ) -> None:
        """
        Index drivers implementing the legacy lineage API can call this method to check get arguments
        """
        if not self._index.supports_external_lineage:
            if include_deriveds:
                raise NotImplementedError(
                    "This index driver only supports the legacy lineage data - include_deriveds not supported."
                )
            if not self._index.supports_external_lineage and (
                include_deriveds or max_depth > 0
            ):
                raise NotImplementedError(
                    "This index driver only supports the legacy lineage data - max_depth not supported."
                )

    @abstractmethod
    def bulk_get(self, ids: Iterable[DSID]) -> Iterable[Dataset]:
        """
        Get multiple datasets by id. (Lineage sources NOT included)

        :param ids: ids to retrieve
        :return: Iterable of Dataset models
        """

    @deprecat(
        reason="The 'get_derived' static method is deprecated in favour of the new lineage API.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def get_derived(self, id_: DSID) -> Iterable[Dataset]:
        """
        Get all datasets derived from a dataset (NOT recursive)

        :param id\\_: dataset id
        """

    @abstractmethod
    def has(self, id_: DSID) -> bool:
        """
        Is this dataset in this index?

        :param id\\_: dataset id
        :return: True if the dataset exists in this index
        """

    @abstractmethod
    def bulk_has(self, ids_: Iterable[DSID]) -> Iterable[bool]:
        """
        Like `has` but operates on a multiple ids.

        For every supplied id check if database contains a dataset with that id.

        :param ids\\_: iterable of dataset ids to check existence in index

        :return: Iterable of bools, true for datasets that exist in index
        """

    @abstractmethod
    def add(
        self,
        dataset: Dataset,
        with_lineage: bool = True,
        archive_less_mature: int | None = None,
    ) -> Dataset:
        """
        Add ``dataset`` to the index. No-op if it is already present.

        :param dataset: Unpersisted dataset model

        :param with_lineage:
           - ``True (default)`` attempt adding lineage datasets if missing
           - ``False`` record lineage relations, but do not attempt
             adding lineage datasets to the db

        :param archive_less_mature: if integer, search for less
            mature versions of the dataset with the int value as a millisecond
            delta in timestamp comparison

        :return: Persisted Dataset model
        """

    @abstractmethod
    def search_product_duplicates(
        self, product: Product, *args: str | Field
    ) -> Iterable[tuple[tuple, Iterable[UUID]]]:
        """
        Find dataset ids who have duplicates of the given set of field names.

        (Search is always restricted by Product)

        Returns a generator returning a tuple containing a namedtuple of
        the values of the supplied fields, and the datasets that match those
        values.

        :param product: The Product to restrict search to
        :param args: field names to identify duplicates over
        """

    @abstractmethod
    def can_update(
        self,
        dataset: Dataset,
        updates_allowed: Mapping[Offset, AllowPolicy] | None = None,
    ) -> tuple[bool, list[Change], list[Change]]:
        """
        Check if dataset can be updated. Return bool,safe_changes,unsafe_changes

        :param dataset: Dataset to update
        :param updates_allowed: Allowed updates
        :return: Tuple of: boolean (can/can't update); safe changes; unsafe changes
        """

    @abstractmethod
    def update(
        self,
        dataset: Dataset,
        updates_allowed: Mapping[Offset, AllowPolicy] | None = None,
        archive_less_mature: int | None = None,
    ) -> Dataset:
        """
        Update dataset metadata and location
        :param dataset: Dataset model with unpersisted updates
        :param updates_allowed: Allowed updates
        :param archive_less_mature: Find and archive less mature datasets with ms delta
        :return: Persisted dataset model
        """

    @abstractmethod
    def archive(self, ids: Iterable[DSID]) -> None:
        """
        Mark datasets as archived

        :param ids: list of dataset ids to archive
        """

    def archive_less_mature(self, ds: Dataset, delta: int | bool = 500) -> None:
        """
        Archive less mature versions of a dataset

        :param ds: dataset to search
        :param delta: millisecond delta for time range.
            If True, default to 500ms. If False, do not find or archive less mature datasets.
            Bool value accepted only for improving backwards compatibility, int preferred.
        """
        less_mature = self.find_less_mature(ds, delta)
        less_mature_ids = (x.id for x in less_mature)

        self.archive(less_mature_ids)
        for lm_ds in less_mature_ids:
            _LOG.info(f"Archived less mature dataset: {lm_ds}")

    def find_less_mature(
        self, ds: Dataset, delta: int | bool = 500
    ) -> Iterable[Dataset]:
        """
        Find less mature versions of a dataset

        :param ds: Dataset to search
        :param delta: millisecond delta for time range.
            If True, default to 500ms. If None or False, do not find or archive less mature datasets.
            Bool value accepted only for improving backwards compatibility, int preferred.
        :return: Iterable of less mature datasets
        """
        if isinstance(delta, bool):
            _LOG.warning("received delta as a boolean value. Int is preferred")
            if delta is True:  # treat True as default
                delta = 500
            else:  # treat False the same as None
                return []
        elif isinstance(delta, int):
            if delta < 0:
                raise ValueError("timedelta must be a positive integer")
        elif delta is None:
            return []
        else:
            raise TypeError("timedelta must be None, a positive integer, or a boolean")

        def check_maturity_information(dataset, props):
            # check that the dataset metadata includes all maturity-related properties
            # passing in the required props to enable greater extensibility should it be needed
            for prop in props:
                if hasattr(dataset.metadata, prop) and (
                    getattr(dataset.metadata, prop) is not None
                ):
                    return
                raise ValueError(
                    f"Dataset {dataset.id} is missing property {prop} required for maturity check"
                )

        check_maturity_information(ds, ["region_code", "time", "dataset_maturity"])

        # 'expand' the date range by `delta` milliseconds to give a bit more leniency in datetime comparison
        expanded_time_range = Range(
            ds.metadata.time.begin - timedelta(milliseconds=delta),
            ds.metadata.time.end + timedelta(milliseconds=delta),
        )
        dupes = self.search(
            product=ds.product.name,
            region_code=ds.metadata.region_code,
            time=expanded_time_range,
        )

        less_mature = []
        for dupe in dupes:
            if dupe.id == ds.id:
                continue

            # only need to check that dupe has dataset maturity, missing/null region_code and time
            # would already have been filtered out during the search query
            check_maturity_information(dupe, ["dataset_maturity"])

            if dupe.metadata.dataset_maturity == ds.metadata.dataset_maturity:
                # Duplicate has the same maturity, which one should be archived is unclear
                raise ValueError(
                    f"A dataset with the same maturity as dataset {ds.id} already exists, "
                    f"with id: {dupe.id}"
                )

            if dupe.metadata.dataset_maturity < ds.metadata.dataset_maturity:
                # Duplicate is more mature than dataset
                # Note that "final" < "nrt"
                raise ValueError(
                    f"A more mature version of dataset {ds.id} already exists, with id: "
                    f"{dupe.id} and maturity: {dupe.metadata.dataset_maturity}"
                )

            less_mature.append(dupe)
        return less_mature

    @abstractmethod
    def restore(self, ids: Iterable[DSID]) -> None:
        """
        Mark datasets as not archived

        :param ids: list of dataset ids to restore
        """

    @abstractmethod
    def purge(
        self, ids: Iterable[DSID], allow_delete_active: bool = False
    ) -> Sequence[DSID]:
        """
        Delete datasets

        :param ids: iterable of dataset ids to purge
        :param allow_delete_active: if false, only archived datasets can be deleted
        :return: list of purged dataset ids
        """

    @abstractmethod
    def get_all_dataset_ids(self, archived: bool) -> Iterable[UUID]:
        """
        Get all dataset IDs based only on archived status

        This will be very slow and inefficient for large databases, and is really
        only intended for small and/or experimental databases.

        :param archived: If true, return all archived datasets, if false, all unarchived datasets
        :return: Iterable of dataset ids
        """

    @deprecat(
        reason="This method has been moved to the Product resource (i.e. dc.index.products.get_field_names)",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    def get_field_names(self, product_name: str | None = None) -> Iterable[str]:
        """
        Get the list of possible search fields for a Product (or all products)

        :param product_name: Name of product, or None for all products
        :return: All possible search field names
        """
        return self._index.products.get_field_names(product_name)

    @deprecat(
        reason="Multiple locations per dataset are now deprecated.  Please use the 'get_location' method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def get_locations(self, id_: DSID) -> Iterable[str]:
        """
        Get (active) storage locations for the given dataset id

        :param id\\_: dataset id
        :return: Storage locations for the dataset
        """

    @abstractmethod
    def get_location(self, id_: DSID) -> str | None:
        """
        Get (active) storage location for the given dataset id

        :param id\\_: dataset id
        :return: Storage location for the dataset - None if no location for the id\\_,
                or if id\\_ not in db.
        """

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def get_archived_locations(self, id_: DSID) -> Iterable[str]:
        """
        Get archived locations for a dataset

        :param id\\_: dataset id
        :return: Archived storage locations for the dataset
        """

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def get_archived_location_times(
        self, id_: DSID
    ) -> Iterable[tuple[str, datetime.datetime]]:
        """
        Get each archived location along with the time it was archived.

        :param id\\_: dataset id
        :return: Archived storage locations, with archive date.
        """

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def add_location(self, id_: DSID, uri: str) -> bool:
        """
        Add a location to the dataset if it doesn't already exist.

        :param id\\_: dataset id
        :param uri: fully qualified uri
        :return: True if a location was added, false if location already existed
        """

    @abstractmethod
    def get_datasets_for_location(
        self, uri: str, mode: str | None = None
    ) -> Iterable[Dataset]:
        """
        Find datasets that exist at the given URI

        :param uri: search uri
        :param mode: 'exact', 'prefix' or None (to guess)
        :return: Matching dataset models
        """

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def remove_location(self, id_: DSID, uri: str) -> bool:
        """
        Remove a location from the dataset if it exists.

        :param id\\_: dataset id
        :param uri: fully qualified uri
        :return: True if location was removed, false if it didn't exist for the database
        """

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be accessible in future releases. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def archive_location(self, id_: DSID, uri: str) -> bool:
        """
        Archive a location of the dataset if it exists and is active.

        :param id\\_: dataset id
        :param uri: fully qualified uri
        :return: True if location was archived
        """

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
        "Archived locations may not be restorable in future releases. "
        "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def restore_location(self, id_: DSID, uri: str) -> bool:
        """
        Un-archive a location of the dataset if it exists.

        :param id\\_: dataset id
        :param uri: fully qualified uri
        :return: True if location was restored
        """

    @abstractmethod
    def search_by_metadata(
        self, metadata: JsonDict, archived: bool | None = False
    ) -> Iterable[Dataset]:
        """
        Perform a search using arbitrary metadata, returning results as Dataset objects.

        Caution - slow! This will usually not use indexes.

        :param metadata: metadata dictionary representing arbitrary search query
        :param archived: False (default): Return active datasets only.
                         None: Include archived and active datasets.
                         True: Return archived datasets only.
        :return: Matching dataset models
        """

    @deprecat(
        deprecated_args={
            "source_filter": {
                "reason": "Filtering by source metadata is deprecated and will be removed in future.",
                "version": "1.9.0",
                "category": ODC2DeprecationWarning,
            }
        }
    )
    @abstractmethod
    def search(
        self,
        limit: int | None = None,
        source_filter: QueryDict | None = None,
        archived: bool | None = False,
        order_by: Iterable[Any] | None = None,
        **query: QueryField,
    ) -> Iterable[Dataset]:
        """
        Perform a search, returning results as Dataset objects.

        Prior to dataccube-1.9.0, search always returned datasets sorted by product.  From 1.9, no ordering
        is guaranteed.  Ordering of results is now unspecified and may vary between index drivers.

        :param limit: Limit number of datasets per product (None/default = unlimited)
        :param source_filter: Filter criteria for sources (None/default = no filtering)
        :param archived: False (default): Return active datasets only.
                         None: Include archived and active datasets.
                         True: Return archived datasets only.
        :param order_by: field or expression by which to order results
        :param query: search query parameters
        :return: Matching datasets
        """

    def get_all_docs_for_product(
        self, product: Product, batch_size: int = 1000
    ) -> Iterable[DatasetTuple]:
        for ds in self.search(product=[product.name]):
            yield DatasetTuple(product, ds.metadata_doc, ds._uris)  # 2.0: ds.uri

    def get_all_docs(
        self, products: Iterable[Product] | None = None, batch_size: int = 1000
    ) -> Iterable[DatasetTuple]:
        """
        Return all datasets in bulk, filtering by product names only. Do not instantiate models.
        Archived datasets and locations are excluded.

        API Note: This API method is not finalised and may be subject to change.

        :param products: Iterable of products used to build the Dataset models.  May come from a different index.
                         Default/None: all products, Products read from the source index.
        :param batch_size: Size of each chunk in the returned iterable (default = 1000)
        :return: Iterable of DatasetTuple named tuples
        """
        # Default implementation calls search
        if products is None:
            products = list(self.products.get_all())
        for product in products:
            yield from self.get_all_docs_for_product(product, batch_size=batch_size)

    def _add_batch(
        self, batch_ds: Iterable[DatasetTuple], cache: dict[str, Any]
    ) -> BatchStatus:
        """
        Add a single "batch" of datasets, provided as DatasetTuples.

        Default implementation is simple loop of add

        API Note: This API method is not finalised and may be subject to change.

        :param batch_ds: An iterable of one batch's worth of DatasetTuples to add
        :return: BatchStatus named tuple.
        """
        b_skipped = 0
        b_added = 0
        b_started = monotonic()
        for ds_tup in batch_ds:
            kwargs = {"uris": ds_tup.uris} if ds_tup.is_legacy else {"uri": ds_tup.uri}
            try:
                ds = Dataset(
                    product=ds_tup.product, metadata_doc=ds_tup.metadata, **kwargs
                )
                self.add(ds, with_lineage=False)
                b_added += 1
            except DocumentMismatchError as e:
                _LOG.warning("%s: Skipping", str(e))
                b_skipped += 1
            except Exception as e:
                _LOG.warning("%s: Skipping", str(e))
                b_skipped += 1
        return BatchStatus(b_added, b_skipped, monotonic() - b_started)

    def _init_bulk_add_cache(self) -> dict[str, Any]:
        """
        Initialise a cache dictionary that may be used to share data between calls to _add_batch()

        API Note: This API method is not finalised and may be subject to change.

        :return: The initialised cache dictionary
        """
        return {}

    def bulk_add(
        self, datasets: Iterable[DatasetTuple], batch_size: int = 1000
    ) -> BatchStatus:
        """
        Add a group of Dataset documents in bulk.

        API Note: This API method is not finalised and may be subject to change.

        :param datasets: An Iterable of DatasetTuples (i.e. as returned by get_all_docs)
        :param batch_size: Number of datasets to add per batch (default 1000)
        :return: BatchStatus named tuple, with `safe` set to None.
        """

        def increment_progress() -> None:
            report_to_user(".", progress_indicator=True)

        n_batches = 0
        n_in_batch = 0
        added = 0
        skipped = 0
        batch = []
        job_started = monotonic()
        inter_batch_cache = self._init_bulk_add_cache()
        for ds_tup in datasets:
            batch.append(ds_tup)
            n_in_batch += 1
            if n_in_batch >= batch_size:
                batch_result = self._add_batch(batch, inter_batch_cache)
                _LOG.info(
                    "Batch %d/%d datasets added in %.2fs: (%.2fdatasets/min)",
                    batch_result.completed,
                    n_in_batch,
                    batch_result.seconds_elapsed,
                    batch_result.completed * 60 / batch_result.seconds_elapsed,
                )
                added += batch_result.completed
                skipped += batch_result.skipped
                batch = []
                n_in_batch = 0
                n_batches += 1
                increment_progress()
        if n_in_batch > 0:
            batch_result = self._add_batch(batch, inter_batch_cache)
            added += batch_result.completed
            skipped += batch_result.skipped
            increment_progress()

        return BatchStatus(added, skipped, monotonic() - job_started)

    @abstractmethod
    def search_by_product(
        self, archived: bool | None = False, **query: QueryField
    ) -> Iterable[tuple[Product, Iterable[Dataset]]]:
        """
        Perform a search, returning datasets grouped by product type.

        :param archived: False (default): Return active datasets only.
                         None: Include archived and active datasets.
                         True: Return archived datasets only.
        :param query: search query parameters
        :return: Matching datasets, grouped by Product
        """

    @abstractmethod
    def search_returning(
        self,
        field_names: Iterable[str] | None = None,
        custom_offsets: Mapping[str, Offset] | None = None,
        limit: int | None = None,
        archived: bool | None = False,
        order_by: Iterable[Any] | None = None,
        **query: QueryField,
    ) -> Generator[NamedTuple]:
        """
        Perform a search, returning only the specified fields.

        This method can be faster than normal search() if you don't need all
        fields of each dataset.

        It also allows for returning rows other than datasets, such as a row
        per uri when requesting field 'uri'.

        :param field_names: Names of desired fields (default = all known search
                            fields, unless custom_offsets is set, see below)
        :param custom_offsets: A dictionary of offsets in the metadata doc for custom
                               fields custom offsets are returned in addition to fields
                               named in field_names. Default is None, field_names only.
                               If field_names is None, and custom_offsets are provided,
                               only the custom offsets are included, over-riding
                               the normal field_names default.
        :param limit: Limit number of dataset (None/default = unlimited)
        :param archived: False (default): Return active datasets only.
                         None: Include archived and active datasets.
                         True: Return archived datasets only.
        :param order_by: a field name, field, function or clause by which to sort
                         output. None is unsorted and may allow faster return of first
                         result depending on the index driver's implementation.
        :param query: search query parameters
        :return: Namedtuple of requested fields, for each matching dataset.
        """

    @abstractmethod
    def count(self, archived: bool | None = False, **query: QueryField) -> int:
        """
        Perform a search, returning count of results.

        :param archived: False (default): Count active datasets only.
                         None: Count archived and active datasets.
                         True: Count archived datasets only.
        :param query: search query parameters
        :return: Count of matching datasets in index
        """

    @abstractmethod
    def count_by_product(
        self, archived: bool | None = False, **query: QueryField
    ) -> Iterable[tuple[Product, int]]:
        """
        Perform a search, returning a count of for each matching product type.

        :param archived: False (default): Count active datasets only.
                         None: Count archived and active datasets.
                         True: Count archived datasets only.
        :param query: search query parameters
        :return: Counts of matching datasets in index, grouped by product.
        """

    @abstractmethod
    def count_by_product_through_time(
        self, period: str, archived: bool | None = False, **query: QueryField
    ) -> Iterable[tuple[Product, Iterable[tuple[Range, int]]]]:
        """
        Perform a search, returning counts for each product grouped in time slices
        of the given period.

        :param period: Time range for each slice: '1 month', '1 day' etc.
        :param archived: False (default): Count active datasets only.
                         None: Count archived and active datasets.
                         True: Count archived datasets only.
        :param query: search query parameters
        :returns: For each matching product type, a list of time ranges and their count.
        """

    @abstractmethod
    def count_product_through_time(
        self, period: str, archived: bool | None = False, **query: QueryField
    ) -> Iterable[tuple[Range, int]]:
        """
        Perform a search, returning counts for a single product grouped in time slices
        of the given period.

        Will raise an error if the search terms match more than one product.

        :param period: Time range for each slice: '1 month', '1 day' etc.
        :param archived: False (default): Count active datasets only.
                         None: Count archived and active datasets.
                         True: Count archived datasets only.
        :param query: search query parameters
        :returns: The product, a list of time ranges and the count of matching datasets.
        """

    @deprecat(
        reason="This method is deprecated and will be removed in 2.0.  "
        "Consider migrating to search_returning()",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    @abstractmethod
    def search_summaries(self, **query: QueryField) -> Iterable[Mapping[str, Any]]:
        """
        Perform a search, returning just the search fields of each dataset.

        :param query: search query parameters
        :return: Mappings of search fields for matching datasets
        """

    @deprecat(
        reason="This method is deprecated and will be removed in 2.0.  "
        "Please use list(dc.index.datasets.search(...)) instead",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    def search_eager(self, **query: QueryField) -> Iterable[Dataset]:
        """
        Perform a search, returning results as Dataset objects.

        :param query: search query parameters
        :return: Fully instantiated list of matching dataset models
        """
        return list(self.search(**query))

    @abstractmethod
    def temporal_extent(
        self, ids: Iterable[DSID]
    ) -> tuple[datetime.datetime, datetime.datetime]:
        """
        Returns the minimum and maximum acquisition time of an iterable of dataset ids.

        Raises KeyError if none of the datasets are in the index

        :param ids: Iterable of dataset ids.
        :return: minimum and maximum acquisition times
        """

    @deprecat(
        reason="This method has been moved to the Product Resource and renamed 'temporal_extent()'",
        version="1.9.0",
        category=ODC2DeprecationWarning,
    )
    def get_product_time_bounds(
        self, product: str | Product
    ) -> tuple[datetime.datetime, datetime.datetime]:
        """
        Returns the minimum and maximum acquisition time of the product.

        :param product: Product of name of product
        :return: minimum and maximum acquisition times
        """
        return self._index.products.temporal_extent(product=product)

    @abstractmethod
    def search_returning_datasets_light(
        self,
        field_names: tuple[str, ...],
        custom_offsets: Mapping[str, Offset] | None = None,
        limit: int | None = None,
        archived: bool | None = False,
        **query: QueryField,
    ) -> Iterable[tuple]:
        """
        This is a dataset search function that returns the results as objects of a dynamically
        generated Dataset class that is a subclass of tuple.

        Only the requested fields will be returned together with related derived attributes as property functions
        similar to the datacube.model.Dataset class. For example, if 'extent'is requested all of
        'crs', 'extent', 'transform', and 'bounds' are available as property functions.

        The field_names can be custom fields in addition to those specified in metadata_type, fixed fields, or
        native fields. The field_names can also be derived fields like 'extent', 'crs', 'transform',
        and 'bounds'. The custom fields require custom offsets of the metadata doc be provided.

        The datasets can be selected based on values of custom fields as long as relevant custom
        offsets are provided. However custom field values are not transformed so must match what is
        stored in the database.

        :param field_names: A tuple of field names that would be returned including derived fields
                            such as extent, crs
        :param custom_offsets: A dictionary of offsets in the metadata doc for custom fields
        :param limit: Number of datasets returned per product.
        :param archived: False (default): Return active datasets only.
                         None: Return archived and active datasets.
                         True: Return archived datasets only.
        :param query: query parameters that will be processed against metadata_types,
                      product definitions and/or dataset table.
        :return: A Dynamically generated DatasetLight (a subclass of namedtuple and possibly with
            property functions).
        """

    @abstractmethod
    def spatial_extent(
        self, ids: Iterable[DSID], crs: CRS = CRS("EPSG:4326")
    ) -> Geometry | None:
        """
        Return the combined spatial extent of the nominated datasets

        Uses spatial index.

        Returns None if no index for the CRS, or if no identified datasets are indexed in the relevant spatial index.
        Result will not include extents of datasets that cannot be validly projected into the CRS.

        :param ids: An iterable of dataset IDs
        :param crs: A CRS (defaults to EPSG:4326)
        :return: The combined spatial extents of the datasets.
        """



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Query datasets.
"""

import csv
import datetime
import shutil
import sys
from collections.abc import Callable, Collection
from functools import partial, singledispatch
from os import terminal_size as t_size
from typing import Any

import click
from sqlalchemy.dialects.postgresql import Range

from datacube.index import Index
from datacube.ui import click as ui
from datacube.ui.click import CLICK_SETTINGS
from datacube.utils.dates import tz_as_utc

PASS_INDEX = ui.pass_index("datacube-search")


def printable_values(d: dict[str, Any]) -> dict[str, str]:
    return {k: printable(v) for k, v in d.items()}


def write_pretty(
    out_f: Any,
    field_names: Collection[str],
    search_results: list[dict[str, Any]],
    terminal_size: t_size = shutil.get_terminal_size(),
) -> None:
    """
    Output in a human-readable text format. Inspired by psql's expanded output.
    """
    terminal_width = terminal_size[0]
    record_num = 1

    field_header_width = max(len(name) for name in field_names)
    field_output_format = "{:<" + str(field_header_width) + "} | {}"

    for result in search_results:
        separator_line = f"-[ {record_num} ]"
        separator_line += "-" * (terminal_width - len(separator_line) - 1)
        click.echo(separator_line, file=out_f)

        for name, value in sorted(result.items()):
            click.echo(field_output_format.format(name, printable(value)), file=out_f)

        record_num += 1


def write_csv(
    out_f: Any, field_names: Collection[str], search_results: list[dict[str, Any]]
) -> None:
    """
    Output as a CSV.
    """
    writer = csv.DictWriter(out_f, tuple(sorted(field_names)))
    writer.writeheader()
    writer.writerows(printable_values(d) for d in search_results)


OUTPUT_FORMATS: dict[str, Callable[[Any, Collection, list[dict[str, Any]]], None]] = {
    "csv": write_csv,
    "pretty": write_pretty,
}


@click.group(help="Search the Data Cube", context_settings=CLICK_SETTINGS)
@ui.global_cli_options
@click.option(
    "-f",
    type=click.Choice(list(OUTPUT_FORMATS)),
    default="pretty",
    show_default=True,
    help="Output format",
)
@click.pass_context
def cli(ctx, f) -> None:
    ctx.obj["write_results"] = partial(OUTPUT_FORMATS[f], sys.stdout)


@cli.command()
@ui.parsed_search_expressions
@PASS_INDEX
@click.pass_context
def datasets(ctx, index: Index, expressions) -> None:
    """
    Search available Datasets
    """
    click.echo(
        "WARNING: The `datacube-search datasets` command is deprecated and will be removed in a future release. "
        "Please use `datacube dataset search` instead.",
        err=True,
    )
    ctx.obj["write_results"](
        sorted(index.products.get_field_names()),
        (tup._asdict() for tup in index.datasets.search_returning(**expressions)),
    )


@cli.command("product-counts")
@click.argument("period", nargs=1)
@ui.parsed_search_expressions
@PASS_INDEX
def product_counts(index: Index, period, expressions) -> None:
    """
    Count product Datasets available by period

    PERIOD: e.g. 1 month, 6 months, 1 year
    """
    click.echo(
        "WARNING: The `datacube-search product-counts` command is deprecated and will be removed in a future release. "
        "Please use `datacube dataset count` instead.",
        err=True,
    )
    for product, series in index.datasets.count_by_product_through_time(
        period, **expressions
    ):
        click.echo(product.name)
        for timerange, count in series:
            formatted_dt = tz_as_utc(timerange[0]).strftime("%Y-%m-%d")
            click.echo(f"    {formatted_dt}: {count}")


@singledispatch
def printable(val) -> str:
    return val


@printable.register(type(None))
def printable_none(val: None) -> str:
    return ""


@printable.register(datetime.datetime)
def printable_dt(val: datetime.datetime) -> str:
    return tz_as_utc(val).isoformat()


@printable.register(Range)
def printable_r(val: Range) -> str:
    if val.lower_inf:
        return printable(val.upper)
    if val.upper_inf:
        return printable(val.lower)

    return f"{printable(val.lower)} to {printable(val.upper)}"


if __name__ == "__main__":
    cli()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2025 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
from types import SimpleNamespace

import numpy as np
import pandas
import pytest
from odc.geo import CRS

from datacube.api.query import GroupBy, Query, query_group_by, solar_day, solar_offset
from datacube.model import Range
from datacube.utils import parse_time


@pytest.fixture
def mock_index():
    from unittest.mock import MagicMock

    idx = MagicMock()
    idx.supports_spatial_indexes = False
    return idx


@pytest.fixture
def test_geom():
    from odc.geo.geom import polygon

    return polygon(
        [
            (-35.0, 110.0),
            (-35.0, 125.0),
            (-45.0, 125.0),
            (-45.0, 110),
            (-35.0, 110),
        ],
        crs="EPSG:4326",
    )


def test_query_kwargs(mock_index) -> None:
    mock_index.products.get_field_names = lambda: {
        "product",
        "lat",
        "sat_path",
        "type_id",
        "time",
        "lon",
        "orbit",
        "instrument",
        "sat_row",
        "platform",
        "metadata_type",
        "gsi",
        "type",
        "id",
    }

    query = Query(index=mock_index, product="ls5_nbar_albers")
    assert str(query)
    assert query.product == "ls5_nbar_albers"
    assert query.search_terms["product"] == "ls5_nbar_albers"

    query = Query(index=mock_index, latitude=(-35, -36), longitude=(148, 149))
    assert query.geopolygon
    assert "lat" in query.search_terms
    assert "lon" in query.search_terms

    query = Query(index=mock_index, latitude=-35, longitude=148)
    assert query.geopolygon
    assert "lat" in query.search_terms
    assert "lon" in query.search_terms

    query = Query(index=mock_index, latitude=-35, longitude=148)

    query = Query(
        index=mock_index, y=(-4174726, -4180011), x=(1515184, 1523263), crs="EPSG:3577"
    )
    assert query.geopolygon
    assert "lat" in query.search_terms
    assert "lon" in query.search_terms

    query = Query(index=mock_index, y=-4174726, x=1515184, crs="EPSG:3577")
    assert query.geopolygon
    assert "lat" in query.search_terms
    assert "lon" in query.search_terms

    query = Query(index=mock_index, y=-4174726, x=1515184, crs="EPSG:3577")
    assert query.geopolygon
    assert "lat" in query.search_terms
    assert "lon" in query.search_terms

    query = Query(index=mock_index, y=-4174726, x=1515184, crs=CRS("EPSG:3577"))
    assert query.geopolygon
    assert "lat" in query.search_terms
    assert "lon" in query.search_terms

    mock_index.supports_spatial_indexes = True
    query = Query(index=mock_index, time="2001")
    assert "time" in query.search

    query = Query(index=mock_index, time=("2001", "2002"))
    assert "time" in query.search

    with pytest.raises(ValueError):
        query = Query(index=mock_index, time=("2001", "2002", "2003"))

    with pytest.raises(ValueError):
        query = Query(index=mock_index, time=["2001", "2002", "2003"])

    with pytest.raises(ValueError):
        Query(
            index=mock_index,
            y=-4174726,
            coordinate_reference_system="WGS84",
            x=1515184,
            crs="EPSG:3577",
        )

    with pytest.raises(LookupError):
        Query(
            index=mock_index,
            y=-4174726,
            x=1515184,
            crs="EPSG:3577",
            made_up_key="NotReal",
        )

    with pytest.raises(LookupError):
        query_group_by(group_by="magic")

    gb = query_group_by("time")
    assert isinstance(gb, GroupBy)
    assert query_group_by(group_by=gb) is gb


def test_query_kwargs_postgis(mock_index, test_geom) -> None:
    mock_index.supports_spatial_indexes = True

    query = Query(index=mock_index, latitude=-35, longitude=148)
    assert query.geopolygon
    assert "lat" not in query.search_terms
    assert "lon" not in query.search_terms
    assert "geopolygon" in query.search_terms

    query = Query(index=mock_index, geopolygon=test_geom)
    assert query.geopolygon
    assert "geopolygon" in query.search_terms


def format_test(start_out, end_out):
    return Range(
        pandas.to_datetime(start_out, utc=True).to_pydatetime(),
        pandas.to_datetime(end_out, utc=True).to_pydatetime(),
    )


testdata = [
    (
        (datetime.datetime(2008, 1, 1), datetime.datetime(2008, 1, 10)),
        format_test("2008-01-01T00:00:00", "2008-01-10T00:00:00.999999"),
    ),
    (
        (datetime.datetime(2008, 1, 1), datetime.datetime(2008, 1, 10, 23, 0, 0)),
        format_test("2008-01-01T00:00:00", "2008-01-10T23:00:00.999999"),
    ),
    (
        (datetime.datetime(2008, 1, 1), datetime.datetime(2008, 1, 10, 23, 59, 40)),
        format_test("2008-01-01T00:00:00", "2008-01-10T23:59:40.999999"),
    ),
    (("2008"), format_test("2008-01-01T00:00:00", "2008-12-31T23:59:59.999999")),
    (
        ("2008", "2008"),
        format_test("2008-01-01T00:00:00", "2008-12-31T23:59:59.999999"),
    ),
    (
        ("2008", "2009"),
        format_test("2008-01-01T00:00:00", "2009-12-31T23:59:59.999999"),
    ),
    (
        ("2008-03", "2009"),
        format_test("2008-03-01T00:00", "2009-12-31T23:59:59.999999"),
    ),
    (
        ("2008-03", "2009-10"),
        format_test("2008-03-01T00:00", "2009-10-31T23:59:59.999999"),
    ),
    (
        ("2008", "2009-10"),
        format_test("2008-01-01T00:00", "2009-10-31T23:59:59.999999"),
    ),
    (
        ("2008-03-03", "2008-11"),
        format_test("2008-03-03T00:00:00", "2008-11-30T23:59:59.999999"),
    ),
    (
        ("2008-11-14", "2008-11-30"),
        format_test("2008-11-14T00:00:00", "2008-11-30T23:59:59.999999"),
    ),
    (
        ("2008-11-14", "2008-11-29"),
        format_test("2008-11-14T00:00:00", "2008-11-29T23:59:59.999999"),
    ),
    (
        ("2008-11-14", "2008-11"),
        format_test("2008-11-14T00:00:00", "2008-11-30T23:59:59.999999"),
    ),
    (
        ("2008-11-14", "2008"),
        format_test("2008-11-14T00:00:00", "2008-12-31T23:59:59.999999"),
    ),
    (("2008-11-14"), format_test("2008-11-14T00:00:00", "2008-11-14T23:59:59.999999")),
    (
        ("2008-11-14", "2009-02-02"),
        format_test("2008-11-14T00:00:00", "2009-02-02T23:59:59.999999"),
    ),
    (
        ("2008-11-14T23:33:57", "2008-11-14 23:33:57"),
        format_test("2008-11-14T23:33:57", "2008-11-14T23:33:57.999999"),
    ),
    (
        ("2008-11-14 23:33", "2008-11-14 23:34"),
        format_test("2008-11-14T23:33:00", "2008-11-14T23:34:59.999999"),
    ),
    (
        ("2008-11-14T23:00:00", "2008-11-14 23:35"),
        format_test("2008-11-14T23:00", "2008-11-14T23:35:59.999999"),
    ),
    (
        ("2008-11-10T11", "2008-11-16 14:01"),
        format_test("2008-11-10T11:00", "2008-11-16T14:01:59.999999"),
    ),
    (
        (datetime.date(1995, 1, 1), datetime.date(1999, 1, 1)),
        format_test("1995-01-01T00:00:00", "1999-01-01T23:59:59.999999"),
    ),
    (
        (datetime.datetime(2008, 1, 1), datetime.datetime(2008, 1, 10, 23, 59, 40)),
        format_test("2008-01-01T00:00:00", "2008-01-10T23:59:40.999999"),
    ),
    (
        (datetime.date(2008, 1, 1)),
        format_test("2008-01-01T00:00:00", "2008-01-01T23:59:59.999999"),
    ),
    (
        (datetime.date(2008, 1, 1), None),
        # Starting the tests before midnight and finishing after will cause test failure
        # since datetime.now() is evaluated before midnight and the test runs after midnight.
        # Put the result behind a function that is evaluated inside the test instead.
        lambda _: format_test(
            "2008-01-01T00:00:00",
            datetime.datetime.now().strftime("%Y-%m-%dT23:59:59.999999"),
        ),
    ),
    (
        (None, "2008"),
        format_test(
            datetime.datetime.fromtimestamp(0).strftime("%Y-%m-%dT%H:%M:%S"),
            "2008-12-31T23:59:59.999999",
        ),
    ),
]


@pytest.mark.parametrize("time_param,expected", testdata)
def test_time_handling(time_param, expected) -> None:
    query = Query(time=time_param)
    assert "time" in query.search_terms
    assert query.search_terms["time"] == (
        expected if isinstance(expected, Range) else expected("now")
    )


def test_time_handling_int() -> None:
    with pytest.raises(TypeError):
        Query(time=2008)


def test_solar_day() -> None:
    _s = SimpleNamespace
    ds = _s(
        center_time=parse_time("1987-05-22 23:07:44.2270250Z"),
        metadata=_s(lon=Range(begin=150.415, end=152.975)),
    )

    # FIXME: Dataset expected.
    assert solar_day(ds) == np.datetime64("1987-05-23", "D")  # type: ignore[arg-type]
    assert solar_day(ds, longitude=0) == np.datetime64("1987-05-22", "D")  # type: ignore[arg-type]

    ds.metadata = _s()

    with pytest.raises(ValueError) as e:
        # FIXME: Dataset expected.
        solar_day(ds)  # type: ignore[arg-type]

    assert "Cannot compute solar_day: dataset is missing spatial info" in str(e.value)

    # Test with Non-UTC timestamp in index.
    ds = _s(
        center_time=parse_time("1997-05-22 22:07:44.2270250-7:00"),
        metadata=_s(lon=Range(begin=-136.615, end=-134.325)),
    )
    # FIXME: Dataset expected.
    assert solar_day(ds) == np.datetime64("1997-05-22", "D")  # type: ignore[arg-type]
    assert solar_day(ds, longitude=0) == np.datetime64("1997-05-23", "D")  # type: ignore[arg-type]


def test_solar_offset() -> None:
    from datetime import timedelta

    from odc.geo.geom import point

    def _hr(t):
        return t.days * 24 + t.seconds / 3600

    def p(lon):
        return point(lon, 0, "epsg:4326")

    assert solar_offset(p(0)) == timedelta(seconds=0)
    assert solar_offset(p(0).to_crs("epsg:3857")) == timedelta(seconds=0)

    assert solar_offset(p(179.9)) == timedelta(hours=12)
    assert _hr(solar_offset(p(-179.9))) == -12.0

    assert solar_offset(p(20), "s") != solar_offset(p(20), "h")
    assert solar_offset(p(20), "s") < solar_offset(p(21), "s")

    _s = SimpleNamespace
    ds = _s(
        center_time=parse_time("1987-05-22 23:07:44.2270250Z"),
        metadata=_s(lon=Range(begin=150.415, end=152.975)),
    )
    # FIXME: Geometry | Dataset expected.
    assert solar_offset(ds) == timedelta(hours=10)  # type: ignore[arg-type]
    ds.metadata = _s()

    with pytest.raises(ValueError):
        # FIXME: Geometry | Dataset expected.
        solar_offset(ds)  # type: ignore[arg-type]


def test_dateline_query_building() -> None:
    lon = Query(
        x=(618300, 849000), y=(-1876800, -1642500), crs="EPSG:32660"
    ).search_terms["lon"]

    assert lon.begin < 180 < lon.end


def test_query_issue_1146() -> None:
    q = Query(k="AB")
    assert q.search["k"] == "AB"


def test_query_multiple_products(mock_index) -> None:
    q = Query(index=mock_index, product=["ls5_nbar_albers", "ls7_nbar_albers"])
    assert q.product == ["ls5_nbar_albers", "ls7_nbar_albers"]
